"""
This module determines which patch files overlap with the given image.

Currently this assumes that the sky projected borders of all images are straight.
"""

import logging
import os
import os.path

import asdf
import gwcs.wcs as wcs
import numpy as np
import spherical_geometry.polygon as sgp
import spherical_geometry.vector as sgv
from spherical_geometry.vector import normalize_vector

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

RAD_TO_ARCSEC = 180.0 / np.pi * 3600.0

PATCH_TABLE = None
CANDIDATE_RADIUS = 0.5  # Radius of circle for which patch centers must lie within
# to be tested for intersection (in degrees)


def load_patch_table(tablepath=None):
    """
    Load the patch table. If no tablepath is supplied the path is obtained
    from the environmental variable PATCH_TABLE_PATH
    """
    global PATCH_TABLE
    if tablepath is None:
        try:
            tablepath = os.environ["PATCH_TABLE_PATH"]
        except KeyError:
            log.error("PATCH_TABLE_PATH environmental variable not found")
            return
    try:
        with asdf.open(tablepath) as af:
            PATCH_TABLE = af.tree["patches"].copy()
    except FileNotFoundError:
        raise FileNotFoundError("Specified patch table file path not found")


def image_coords_to_vec(image_corners):
    """
    This routine can handle the corners in both organizations, whether
    a sequence of ra, dec pairs or a list of ra coordinates, followed
    by a list of dec coordinates. So long as either form can be converted
    to a numpy array it is ok. The spherical geometry routines expect
    the latter organization, and the returned value will be of that
    organization.
    """
    image_corners = np.array(image_corners)
    if image_corners.shape == (4, 2):
        image_corners = image_corners.transpose()
    # Convert all celestial coordinates to cartesion coordinates.
    vec_im_corners = np.array(sgv.lonlat_to_vector(image_corners[0], image_corners[1]))
    return vec_im_corners


def find_patch_matches(image_corners, image_shape=None):
    """Find patches that the image overlaps with

    Parameters
    ----------
    image_corners : Either a squence of 4 (ra, dec) pairs, or
        equivalent 2-d numpy array, or
        a GWCS instance. The instance must have either the bounding_box or
        pixel_shape attribute defined, or the following image_shape argument
        must be supplied

    image_shape : image shape to be used if a GWCS instance is supplied
        and does not contain a value for either the bounding_box or
        pixel_shape attributes. Default value is None.

    Returns
    -------
    A sequence of the indices of all patches that overlap the supplied image
    (in the referenced patch table). The indices may be used to obtain all
    necessary information about the patches.
    """

    if PATCH_TABLE is None:
        raise RuntimeError("No patch table has been loaded")
    if isinstance(image_corners, wcs.WCS):
        iwcs = image_corners
        # Now must find size of correspinding image, with three possible
        # sources of that information.
        if (
            (not hasattr(iwcs, "bounding_box") or iwcs.bounding_box is None)
            and (not hasattr(iwcs, "pixel_shape") or iwcs.pixel_shape is None)
            and image_shape is None
        ):
            raise ValueError(
                "Use of a wcs object requires at least one of the bounding_box"
                " or pixel_shape attributes be set to the image shape or that the"
                "image_shape argument be set"
            )
        if image_shape is not None:
            pass
        else:
            # Both bounding_box and pixel_shape are in x, y order contrary to
            # numpy convention.
            if hasattr(iwcs, "bounding_box") and iwcs.bounding_box is not None:
                # Presumes that the bounding_box matches image array boundaries
                bbintervals = iwcs.bounding_box.intervals
                # This compensates for the half pixel adjustment in the general code.
                image_shape = (bbintervals[1].upper + 0.5, bbintervals[0].upper + 0.5)
            elif hasattr(iwcs, "pixel_shape") and iwcs.pixel_shape is not None:
                image_shape = (iwcs.pixel_shape[1], iwcs.pixel_shape[0])
        # Compute the image corners ra, dec from the wcs
        (cxm, cxp), (cym, cyp) = (
            (-0.5, image_shape[1] - 0.5),
            (-0.5, image_shape[0] - 0.5),
        )
        image_corners = (iwcs(cxp, cyp), iwcs(cxm, cyp), iwcs(cxm, cym), iwcs(cxp, cym))
    ptab = PATCH_TABLE
    ra = ptab[:]["ra_center"]
    dec = ptab[:]["dec_center"]
    # # Convert all celestial coordinates to cartesion coordinates.
    vec_centers = np.array(sgv.lonlat_to_vector(ra, dec)).transpose()
    # # Organize corners into two ra, dec lists
    # imra = np.array([point[0] for point in image_corners])
    # imdec = np.array([point[1] for point in image_corners])
    vec_im_corners = image_coords_to_vec(image_corners)
    # Approximate center of image by averaging corner vectors
    im_center = normalize_vector(vec_im_corners.mean(axis=1))
    # Compute difference vector between image center and patch centers
    diff = vec_centers - im_center
    dist = np.sqrt((diff**2).sum(axis=1)) * 180 / np.pi
    match = np.where(dist < 0.5)
    ncandidates = len(match[0])
    # Now see which of these that are close actually overlap the supplied image.
    # (Is it necessary to check that the corners are in a sensible order?)
    # All the corner coordinates are returned as arrays.
    mra1 = ptab[match]["ra_corn1"]
    mra2 = ptab[match]["ra_corn2"]
    mra3 = ptab[match]["ra_corn3"]
    mra4 = ptab[match]["ra_corn4"]
    mdec1 = ptab[match]["dec_corn1"]
    mdec2 = ptab[match]["dec_corn2"]
    mdec3 = ptab[match]["dec_corn3"]
    mdec4 = ptab[match]["dec_corn4"]
    mcenters = vec_centers[match]
    mra = np.vstack([mra1, mra2, mra3, mra4, mra1])
    mdec = np.vstack([mdec1, mdec2, mdec3, mdec4, mdec1])
    points = np.array(sgv.lonlat_to_vector(mra, mdec))
    # Create polygon for supplied image_corners
    pvec_im_corners = np.empty((5, 3))
    pvec_im_corners[:4] = vec_im_corners.transpose()
    pvec_im_corners[4] = pvec_im_corners[0]
    impoly = sgp.SingleSphericalPolygon(pvec_im_corners, im_center)
    realmatch = []
    for i in range(ncandidates):
        # print(i)
        cellpoints = points[:, :, i].transpose()
        cellcenter = mcenters[i]
        cellpoly = sgp.SingleSphericalPolygon(cellpoints, cellcenter)
        if impoly.intersects_poly(cellpoly):
            # print(f"candidate {i} intersects")
            realmatch.append(i)
    return match[0][realmatch], match[0]


def get_cartesian_corners(patch):
    """
    Cons] truct a the vertex coordinates for a patch from a patch definition suitable
    for plotting the defined region (ra coordinates, dec coordinates). It returns
    coordinates in the cartesian system so that it avoids the distortions in plots
    due to areas approaching the poles.
    """
    dec_corners = np.array(
        (
            patch["dec_corn1"],
            patch["dec_corn2"],
            patch["dec_corn3"],
            patch["dec_corn4"],
            patch["dec_corn1"],
        )
    )
    ra_corners = np.array(
        (
            patch["ra_corn1"],
            patch["ra_corn2"],
            patch["ra_corn3"],
            patch["ra_corn4"],
            patch["ra_corn1"],
        )
    )
    vec_corners = sgv.lonlat_to_vector(ra_corners, dec_corners)
    return vec_corners


def find_closest_tangent_point(patches, image_corners):
    """
    Out of all listed patches, find the closest tangent point to the center
    coordinate of the image.
    """
    # To deal with the corner case, it is necessary to use spherical_geometry
    # to average the the corners properly.
    # Convert image corners to unit vectors.
    vec_im_corners = image_coords_to_vec(image_corners)
    im_center = np.array(normalize_vector(vec_im_corners.mean(axis=1)))
    tangent_point_set = set()
    patch_tangent_points = [
        (patch["ra_projection_center"], patch["dec_projection_center"])
        for patch in patches
    ]
    for tangent_point in patch_tangent_points:
        tangent_point_set.add(tangent_point)
    unique_tangent_points = list(tangent_point_set)
    # Compute distance for each tangent point from im_center
    dist = [
        ((im_center - np.array(sgv.lonlat_to_vector(*tangent_point))) ** 2).sum()
        for tangent_point in unique_tangent_points
    ]
    sorted_dist_indices = sorted(zip(dist, range(len(dist))))
    sorted_tangent_points = [
        unique_tangent_points[sorted_dist[1]] for sorted_dist in sorted_dist_indices
    ]
    closest_tangent_point = np.array(sgv.lonlat_to_vector(*sorted_tangent_points[0]))
    # Now associate index of sorted_tangent_points with that of all patches
    patch_tp_id = []
    for patch_tp in patch_tangent_points:
        for i, tp in enumerate(sorted_tangent_points):
            if tp == patch_tp:
                patch_tp_id.append(i)
    return closest_tangent_point, patch_tp_id


def veccoords_to_tangent_plane(vertices, tangent_point_vec):
    """
    Convert the spherical geometry vectors to tangent plane coordinates
    in arcseconds. This algorithm is not precise, but should be good
    enough for now (and besides, the goal here is visualizaion, not
    ultra-precision). This also breaks down numerically very near the
    poles.
    """
    # First compute the tangent plane axis vectors.
    x_axis = normalize_vector(np.cross([0, 0, 1], tangent_point_vec))
    y_axis = normalize_vector(
        np.array([0, 0, 1])
        - np.array(tangent_point_vec)
        * np.dot(np.array([0, 0, 1]), np.array(tangent_point_vec))
    )
    avertices = np.vstack(vertices)
    x_coords = np.dot(x_axis, avertices) * RAD_TO_ARCSEC
    y_coords = np.dot(y_axis, avertices) * RAD_TO_ARCSEC
    return x_coords, y_coords


load_patch_table()



"""Tests for the DQ Init module and DMS 25 and DMS 26 requirements"""

import os

import pytest
import roman_datamodels as rdm
from metrics_logger.decorators import metrics_logger

from romancal.step import DQInitStep
from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@metrics_logger("DMS25")
@pytest.mark.bigdata
def test_dq_init_image_step(rtdata, ignore_asdf_paths):
    """DMS25 Test: Testing retrieval of best ref file for image data,
    and creation of a ramp file with CRDS selected mask file applied."""

    input_file = "r0000101001001001001_01101_0001_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/image/{input_file}")
    rtdata.input = input_file

    # Test CRDS
    step = DQInitStep()
    model = rdm.open(rtdata.input)
    step.log.info(
        "DMS25 MSG: Testing retrieval of best "
        "ref file for image data, "
        "Success is creation of a ramp file with CRDS selected "
        "mask file applied."
    )

    step.log.info(f'DMS25 MSG: First data file: {rtdata.input.rsplit("/", 1)[1]}')
    ref_file_path = step.get_reference_file(model, "mask")
    step.log.info(
        f'DMS25 MSG: CRDS matched mask file: {ref_file_path.rsplit("/", 1)[1]}'
    )
    ref_file_name = os.path.split(ref_file_path)[-1]

    assert "roman_wfi_mask" in ref_file_name

    # Test DQInitStep
    output = "r0000101001001001001_01101_0001_WFI01_dqinit.asdf"
    rtdata.output = output
    args = ["romancal.step.DQInitStep", rtdata.input]
    step.log.info(
        "DMS25 MSG: Running data quality initialization step."
        " The first ERROR is expected, due to extra CRDS parameters"
        " not having been implemented yet."
    )
    RomanStep.from_cmdline(args)
    ramp_out = rdm.open(rtdata.output)
    step.log.info(
        "DMS25 MSG: Does ramp data contain pixeldq from mask file? :"
        f' {("roman.pixeldq" in ramp_out.to_flat_dict())}'
    )
    assert "roman.pixeldq" in ramp_out.to_flat_dict()

    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    step.log.info(
        "DMS25 MSG: Was the proper data quality array initialized"
        " for the ramp data produced? : "
        f"{diff.identical}"
    )
    assert diff.identical, diff.report()


@metrics_logger("DMS25")
@pytest.mark.bigdata
def test_dq_init_grism_step(rtdata, ignore_asdf_paths):
    """DMS25 Test: Testing retrieval of best ref file for grism data,
    and creation of a ramp file with CRDS selected mask file applied."""

    input_file = "r0000201001001001001_01101_0001_WFI01_uncal.asdf"
    rtdata.get_data(f"WFI/grism/{input_file}")
    rtdata.input = input_file

    # Test CRDS
    step = DQInitStep()
    model = rdm.open(rtdata.input)
    step.log.info(
        "DMS25 MSG: Testing retrieval of best "
        "ref file for grism data, "
        "Success is creation of a ramp file with CRDS selected "
        "mask file applied."
    )

    step.log.info(f'DMS25 MSG: First data file: {rtdata.input.rsplit("/", 1)[1]}')
    ref_file_path = step.get_reference_file(model, "mask")
    step.log.info(
        f'DMS25 MSG: CRDS matched mask file: {ref_file_path.rsplit("/", 1)[1]}'
    )
    ref_file_name = os.path.split(ref_file_path)[-1]

    assert "roman_wfi_mask" in ref_file_name

    # Test DQInitStep
    output = "r0000201001001001001_01101_0001_WFI01_dqinit.asdf"
    rtdata.output = output
    args = ["romancal.step.DQInitStep", rtdata.input]
    step.log.info(
        "DMS25 MSG: Running data quality initialization step."
        "The first ERROR is expected, due to extra CRDS parameters "
        "not having been implemented yet."
    )
    RomanStep.from_cmdline(args)
    ramp_out = rdm.open(rtdata.output)
    step.log.info(
        "DMS25 MSG: Does ramp data contain pixeldq from mask file? :"
        f' {("roman.pixeldq" in ramp_out.to_flat_dict())}'
    )
    assert "roman.pixeldq" in ramp_out.to_flat_dict()

    rtdata.get_truth(f"truth/WFI/grism/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    step.log.info(
        "DMS25 MSG: Was proper data quality initialized "
        "ramp data produced? : "
        f"{diff.identical}"
    )
    assert diff.identical, diff.report()



""" Roman tests for the High Level Pipeline """

import os

import pytest
import roman_datamodels as rdm
from metrics_logger.decorators import metrics_logger

from romancal.pipeline.highlevel_pipeline import HighLevelPipeline

from .regtestdata import compare_asdf


def passfail(bool_expr):
    """set pass fail"""
    if bool_expr:
        return "Pass"
    return "Fail"


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS356")
def test_level3_hlp_pipeline(rtdata, ignore_asdf_paths):
    """Tests for level 3 processing requirements DMS356"""

    cal_files = [
        "WFI/image/r0000101001001001001_01101_0001_WFI01_cal.asdf",
        "WFI/image/r0000101001001001001_01101_0002_WFI01_cal.asdf",
        "WFI/image/r0000101001001001001_01101_0003_WFI01_cal.asdf",
    ]

    for cal_file in cal_files:
        rtdata.get_data(cal_file)

    input_asn = "L3_regtest_asn.json"
    rtdata.get_data(f"WFI/image/{input_asn}")
    rtdata.input = input_asn

    # Test Pipeline
    output = "r0099101001001001001_F158_visit_i2d.asdf"
    rtdata.output = output
    args = [
        "--disable-crds-steppars",
        "roman_hlp",
        rtdata.input,
    ]
    HighLevelPipeline.from_cmdline(args)
    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()

    # Generate thumbnail image
    input_file = "r0099101001001001001_F158_visit_i2d.asdf"
    thumbnail_file = "r0099101001001001001_F158_visit_thumb.png"
    input_file = "r0099101001001001001_F158_visit_i2d.asdf"
    thumbnail_file = "r0099101001001001001_F158_visit_thumb.png"

    preview_cmd = f"stpreview to {input_file} {thumbnail_file} 256 256 roman"
    os.system(preview_cmd)  # nosec

    # Generate preview image
    input_file = "r0099101001001001001_F158_visit_i2d.asdf"
    preview_file = "r0099101001001001001_F158_visit_preview.png"
    preview_cmd = f"stpreview to {input_file} {preview_file} 1080 1080 roman"
    os.system(preview_cmd)  # nosec

    # Perform DMS tests
    # Initial prep
    model = rdm.open(rtdata.output, lazy_load=False)
    pipeline = HighLevelPipeline()

    # DMS356 result is an ImageModel
    pipeline.log.info(
        "DMS356 MSG: Testing that result is a Level 3 mosaic model......."
        + passfail(isinstance(model, rdm.datamodels.MosaicModel))
    )

    # DMS356 Test that skymatch step is complete
    pipeline.log.info(
        "Status of the step:             skymatch    "
        + str(model.meta.cal_step.skymatch)
    )
    # DMS356 Test that the thumbnail image exists
    pipeline.log.info(
        "Status of the step:             thumbnail image    "
        + passfail(os.path.isfile(thumbnail_file))
    )
    # DMS356 Test that the preview image exists
    pipeline.log.info(
        "Status of the step:             preview image    "
        + passfail(os.path.isfile(preview_file))
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of skymatch in the Level 3  output......."
        + passfail(model.meta.cal_step.skymatch == "COMPLETE")
    )
    assert model.meta.cal_step.skymatch == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             skymatch    "
        + str(model.meta.cal_step.skymatch)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of outlier detection in the Level 3 image output......."
        + passfail(model.meta.cal_step.outlier_detection == "COMPLETE")
    )
    assert model.meta.cal_step.outlier_detection == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             outlier_detection    "
        + str(model.meta.cal_step.outlier_detection)
    )
    assert model.meta.cal_step.resample == "COMPLETE"
    pipeline.log.info(
        "Status of the step:             resample          "
        + str(model.meta.cal_step.resample)
    )
    pipeline.log.info(
        "DMS86 MSG: Testing completion of resample in the Level 3 image output......."
        + passfail(model.meta.cal_step.resample == "COMPLETE")
    )



"""Test basic generate operations"""

from romancal.associations import (
    AssociationPool,
    AssociationRegistry,
    generate,
    load_asn,
)
from romancal.associations.tests.helpers import t_path


def test_simple():
    """Test generate on simple registry"""
    registry = AssociationRegistry(
        [t_path("data/rules_basic.py")], include_default=False
    )
    pool = AssociationPool()
    pool["value"] = ["row1", "row2"]

    asns = generate(pool, registry)
    assert len(asns) == 1
    assert len(asns[0]["members"]) == 2


def test_unserialize():
    """Test basic unserializing"""
    asn_file = t_path("data/asn_mosaic.json")
    with open(asn_file) as asn_fp:
        asn = load_asn(asn_fp)
    assert isinstance(asn, dict)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from roman_datamodels.datamodels import RampModel, RefpixRefModel

from .data import Coefficients, StandardView

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def run_steps(
    datamodel: RampModel,
    refs: RefpixRefModel,
    remove_offset: bool = True,
    remove_trends: bool = True,
    cosine_interpolate: bool = True,
    fft_interpolate: bool = True,
) -> RampModel:
    """
    Organize the steps to run the reference pixel correction.
    """

    # Read in the data from the datamodels
    log.debug("Reading data from datamodel into single array")
    coeffs = Coefficients.from_ref(refs)
    standard = StandardView.from_datamodel(datamodel)

    # Remove offset from the data
    if remove_offset:
        standard = standard.remove_offset()
        log.debug("Removed the general offset from data, to be re-applied later.")

    # Convert to channel view
    channel = standard.channels

    # Remove the boundary trends
    if remove_trends:
        channel = channel.remove_trends()
        log.debug("Removed boundary trends (in time) from data.")

    # Cosine interpolate the the data
    if cosine_interpolate:
        channel = channel.cosine_interpolate()
        log.debug("Cosine interpolated the reference pixels.")

    # FFT interpolate the data
    if fft_interpolate:
        channel = channel.fft_interpolate()
        log.debug("FFT interpolated the reference pixel pads.")

    # Perform the reference pixel correction
    standard = channel.apply_correction(coeffs)
    log.debug("Applied reference pixel correction")

    # Re-apply the offset (if necessary)
    standard.apply_offset()
    log.debug("Re-applied the general offset (if removed) to the data.")

    # Write the data back to the datamodel
    standard.update(datamodel)
    log.debug("Updated the datamodel with the corrected data.")

    return datamodel



"""Association Candidate Identifier"""

import re
from ast import literal_eval

from .counter import Counter

__all__ = ["ACID"]

# Start of the discovered association ids.
_DISCOVERED_ID_START = 3001


class ACID:
    """Association Candidate Identifer

    Parameters
    ----------
    input : str or 2-tuple
        The string representation or 2-tuple containing the
        candidate ID and TYPE. The string should be itself
        the 2-tuple representation when evaluated. The 2-tuple
        will have the form:
            (id, type)

    Attributes
    ----------
    id : str
        The id number.

    type : str
        The type of candidate. Some, but not all
        possibilities include 'OBSERVATION',
        'MOSAIC', 'DISCOVERED'

    __str__ : str
        The DMS-specified string representation of
        a candidate identifier. Possibilities include
        'oXXX' for OBSERVATION, 'c1XXX' for 'MOSAIC' or
        other PPS-defined candidates, and 'a3XXX' for
        'DISCOVERED' associations.
    """

    def __init__(self, input):
        try:
            self.id, self.type = literal_eval(input)
        except (ValueError, SyntaxError):
            self.id, self.type = input

    def __str__(self):
        return self.id


class ACIDMixin:
    """Enable ACID for rules"""

    def __init__(self, *args, **kwargs):
        # Initialize discovered association ID
        self.discovered_id = Counter(_DISCOVERED_ID_START)

        super().__init__(*args, **kwargs)

    def acid_from_constraints(self):
        """Determine ACID from constraints"""
        for constraint in self.constraints:
            if getattr(constraint, "is_acid", False):
                value = re.sub("\\\\", "", "-".join(constraint.found_values))
                try:
                    acid = ACID(value)
                except ValueError:
                    pass
                else:
                    break
        else:
            id = f"a{self.discovered_id.value:0>3}"
            acid = ACID((id, "DISCOVERED"))

        return acid



"""Regression tests for the Reference Pixel Correction step"""

import pytest

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_refpix_step(rtdata, ignore_asdf_paths):
    # I have no idea what this is supposed to be
    input_datafile = "r0000101001001001001_01101_0001_WFI01_saturation.asdf"
    rtdata.get_data(f"WFI/image/{input_datafile}")
    rtdata.input = input_datafile

    args = ["romancal.step.RefPixStep", rtdata.input]
    RomanStep.from_cmdline(args)

    # Again I have no idea here
    output = "r0000101001001001001_01101_0001_WFI01_refpix.asdf"
    rtdata.output = output
    rtdata.get_truth(f"truth/WFI/image/{output}")

    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert diff.identical, diff.report()



"""Constraint Tests"""

import pytest

from romancal.associations.lib.constraint import (
    Constraint,
    SimpleConstraint,
    SimpleConstraintABC,
)


def test_sc_dup_names():
    """Test that SimpleConstraint returns an empty dict"""
    sc = SimpleConstraint(name="sc_name")
    dups = sc.dup_names
    assert not len(dups)


class TestDupNames:
    """Test duplicate names in Constraint"""

    # Sub constraints
    sc1 = SimpleConstraint(name="sc1")
    sc2 = SimpleConstraint(name="sc2")
    c1 = Constraint([sc1, sc2], name="c1")
    c2 = Constraint([sc1, sc1], name="c2")
    c3 = Constraint([sc1, sc2], name="sc1")

    @pytest.mark.parametrize(
        "constraints, expected",
        [
            ([sc1], {}),
            ([sc1, sc2], {}),
            ([sc1, sc1], {"sc1": [sc1, sc1]}),
            ([c1], {}),
            ([c2], {"sc1": [sc1, sc1]}),
            ([c3], {"sc1": [sc1, c3]}),
        ],
    )
    def test_dups(self, constraints, expected):
        c = Constraint(constraints)
        dups = c.dup_names
        assert set(dups.keys()) == set(expected.keys())
        for name, constraints in dups.items():
            assert set(dups[name]) == set(expected[name])


def test_sc_get_all_attr():
    """Get attribute value of a simple constraint"""
    name = "my_sc"
    sc = SimpleConstraint(name=name, value="my_value")
    assert sc.get_all_attr("name") == [(sc, name)]


def test_constraint_get_all_attr():
    """Get attribute value of all constraints in a constraint"""
    names = ["sc1", "sc2"]
    constraints = [SimpleConstraint(name=name) for name in names]
    c = Constraint(constraints, name="c1")

    expected = [(constraint, constraint.name) for constraint in constraints]
    expected.append((c, "c1"))
    result = c.get_all_attr("name")
    assert set(result) == set(expected)


def test_simpleconstraint_reprocess_match():
    """Test options for reprocessing"""
    sc = SimpleConstraint(value="my_value", reprocess_on_match=True)
    match, reprocess = sc.check_and_set("my_value")
    assert match
    assert len(reprocess)


def test_simpleconstraint_reprocess_nomatch():
    """Test options for reprocessing"""
    sc = SimpleConstraint(value="my_value", reprocess_on_fail=True)
    match, reprocess = sc.check_and_set("bad_value")
    assert not match
    assert len(reprocess)


def test_constraint_reprocess_match():
    """Test options for reprocessing"""
    sc = SimpleConstraint(value="my_value")
    c = Constraint([sc], reprocess_on_match=True)
    match, reprocess = c.check_and_set("my_value")
    assert match
    assert len(reprocess)


def test_constraint_reprocess_nomatch():
    """Test options for reprocessing"""
    sc = SimpleConstraint(value="my_value")
    c = Constraint([sc], reprocess_on_fail=True)
    match, reprocess = c.check_and_set("bad_value")
    assert not match
    assert len(reprocess)


def test_abc():
    """Test ABC itself"""
    with pytest.raises(TypeError):
        SimpleConstraintABC()


def test_simpleconstraint():
    """Test initialization"""

    # Basic initialization
    c = SimpleConstraint()
    assert c.value is None
    assert c.force_unique
    assert c.test == c.eq

    # Parameter initialization
    c = SimpleConstraint(value="my_value")
    assert c.value == "my_value"

    # Dict initialization
    c = SimpleConstraint({"value": "my_value"})
    assert c.value == "my_value"


def test_simpleconstraint_checkset():
    """Test check_and_set"""

    # Check and set.
    c = SimpleConstraint()
    match, reprocess = c.check_and_set("my_value")
    assert match
    assert c.value == "my_value"
    assert len(reprocess) == 0

    # Non-match
    c = SimpleConstraint(value="my_value")
    match, reprocess = c.check_and_set("bad_value")
    assert not match
    assert c.value == "my_value"
    assert len(reprocess) == 0

    # Don't force unique
    c = SimpleConstraint(force_unique=False)
    match, reprocess = c.check_and_set("my_value")
    assert match
    assert c.value is None
    assert len(reprocess) == 0


def test_constraint_default():
    """Test constraint operations"""

    sc1 = SimpleConstraint()
    sc2 = SimpleConstraint()
    c = Constraint([sc1, sc2])
    match, reprocess = c.check_and_set("my_value")
    assert match
    for constraint in c.constraints:
        assert constraint.value == "my_value"


def test_invalid_init():
    with pytest.raises(TypeError):
        Constraint("bad init")


def test_constraint_all():
    """Test the all operation"""

    sc1 = SimpleConstraint(value="value_1")
    sc2 = SimpleConstraint(value="value_2")
    c = Constraint([sc1, sc2])
    match, reprocess = c.check_and_set("value_1")
    assert not match


def test_constraint_any_basic():
    """Test the all operation"""

    sc1 = SimpleConstraint(value="value_1")
    sc2 = SimpleConstraint(value="value_2")
    c = Constraint([sc1, sc2], reduce=Constraint.any)
    match, reprocess = c.check_and_set("value_1")
    assert match
    match, reprocess = c.check_and_set("value_2")
    assert match
    match, reprocess = c.check_and_set("value_3")
    assert not match


def test_constraint_any_remember():
    """Ensure that any doesn't forget other or propositions"""

    sc1 = SimpleConstraint(value="value_1")
    sc2 = SimpleConstraint(value="value_2")
    c = Constraint([sc1, sc2], reduce=Constraint.any)
    match, reprocess = c.check_and_set("value_1")
    assert match
    match, reprocess = c.check_and_set("value_2")
    assert match
    match, reprocess = c.check_and_set("value_1")
    assert match
    match, reprocess = c.check_and_set("value_3")
    assert not match


def test_iteration():
    """Test various iterations"""
    sc = SimpleConstraint()
    for idx in sc:
        assert isinstance(idx, SimpleConstraint)

    c = Constraint([sc, sc])
    count = 0
    for idx in c:
        assert isinstance(idx, SimpleConstraint)
        count += 1
    assert count == 2

    c = Constraint([Constraint([sc, sc]), Constraint([sc, sc])])
    count = 0
    for idx in c:
        assert isinstance(idx, SimpleConstraint)
        count += 1
    assert count == 4  # Not 6


def test_name_index():
    """Test for name indexing"""
    sc1 = SimpleConstraint(name="sc1", value="value1")
    sc2 = SimpleConstraint(name="sc2", value="value2")
    c1 = Constraint([sc1, sc2])
    assert c1["sc1"].value
    assert c1["sc2"].value

    sc3 = SimpleConstraint(name="sc3", value="value3")
    sc4 = SimpleConstraint(name="sc4", value="value4")
    c2 = Constraint([sc3, sc4, c1])
    assert c2["sc1"].value
    assert c2["sc2"].value
    assert c2["sc3"].value
    assert c2["sc4"].value

    with pytest.raises(KeyError):
        c2["nonexistent"].value

    with pytest.raises(AttributeError):
        c2["sc1"].nonexistent


def test_copy():
    sc1 = SimpleConstraint(name="sc1")
    sc1_copy = sc1.copy()
    assert id(sc1) != id(sc1_copy)
    sc1.check_and_set("value1")
    assert sc1.value == "value1"
    assert sc1_copy.value is None
    sc1_copy.check_and_set("value2")
    assert sc1_copy.value == "value2"
    assert sc1.value == "value1"


@pytest.mark.parametrize(
    "klass, expected",
    [(SimpleConstraint, "SimpleConstraint:myname"), (Constraint, "Constraint:myname")],
)
def test_id(klass, expected):
    """Test constraint ID"""
    c = klass(name="myname")

    assert c.id == expected



from unittest.mock import MagicMock

import numpy as np
import pytest

from ..gwcs_drizzle import GWCSDrizzle


@pytest.mark.parametrize(
    "product, outwcs, wt_scl, pixfrac, kernel, fillval, expected_exception",
    [
        (
            MagicMock(
                data=np.array([[1, 2], [3, 4]]),
                weight=np.array([[0.5, 0.5], [0.5, 0.5]]),
                meta=MagicMock(
                    wcs="WCS", resample=MagicMock(product_exposure_time=1.0)
                ),
                context=np.zeros((1, 2, 2), dtype=np.int32),
            ),
            None,
            None,
            1.0,
            "square",
            "INDEF",
            None,
        ),
        (
            MagicMock(
                data=np.array([[1, 2], [3, 4]]),
                weight=np.array([[1, 1], [1, 1]]),
                meta=MagicMock(
                    wcs="WCS", resample=MagicMock(product_exposure_time=2.0)
                ),
                context=np.zeros((1, 2, 2), dtype=np.int32),
            ),
            "CustomWCS",
            "exptime",
            0.5,
            "gaussian",
            "0",
            None,
        ),
        (
            MagicMock(
                data=np.array([[0, 1], [2, 3]]),
                weight=np.array([[1, 0], [0, 1]]),
                meta=MagicMock(
                    wcs="WCS", resample=MagicMock(product_exposure_time=0.5)
                ),
                context=np.zeros((1, 2, 2), dtype=np.int32),
            ),
            None,
            "expsq",
            0.8,
            "lanczos3",
            "NaN",
            None,
        ),
    ],
)
def test_gwcs_drizzle_init_happy_path(
    product, outwcs, wt_scl, pixfrac, kernel, fillval, expected_exception
):
    if expected_exception:
        with pytest.raises(expected_exception):
            GWCSDrizzle(product, outwcs, wt_scl, pixfrac, kernel, fillval)
    else:
        drizzle = GWCSDrizzle(product, outwcs, wt_scl, pixfrac, kernel, fillval)

    if not expected_exception:
        assert drizzle.outwcs == (outwcs or product.meta.wcs)
        assert drizzle.wt_scl == ("" if wt_scl is None else wt_scl)
        assert drizzle.pixfrac == pixfrac
        assert drizzle.kernel == kernel
        assert drizzle.fillval == fillval


@pytest.mark.parametrize(
    "pixfrac, kernel, expected_pixfrac, expected_kernel",
    [
        (0, "square", 0, "square"),
        (
            2.0,
            "nonexistent",
            2.0,
            "nonexistent",
        ),  # Invalid kernel, but no validation in __init__
    ],
)
def test_gwcs_drizzle_init_edge_cases(
    pixfrac, kernel, expected_pixfrac, expected_kernel
):
    # Arrange
    product = MagicMock(
        data=np.array([[1, 2], [3, 4]]),
        weight=np.array([[1, 1], [1, 1]]),
        meta=MagicMock(wcs="WCS", resample=MagicMock(product_exposure_time=1.0)),
        context=np.zeros((1, 2, 2), dtype=np.int32),
    )

    # Act
    drizzle = GWCSDrizzle(product, pixfrac=pixfrac, kernel=kernel)

    # Assert
    assert drizzle.pixfrac == expected_pixfrac
    assert drizzle.kernel == expected_kernel



"""
Association Pools
"""

from collections import UserDict

from astropy.io.ascii import convert_numpy
from astropy.table import Table

__all__ = ["AssociationPool"]

DEFAULT_DELIMITER = "|"
DEFAULT_FORMAT = "ascii"


class AssociationPool(Table):
    """Association Pool

    An ``AssociationPool`` is essentially an astropy Table with the
    following default behaviors:

    - ASCII tables with a default delimiter of `|`
    - All values are read in as strings
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.meta["pool_file"] = self.meta.get("pool_file", "in-memory")

    @classmethod
    def read(
        cls, filename, delimiter=DEFAULT_DELIMITER, format=DEFAULT_FORMAT, **kwargs
    ):
        """Read in a Pool file

        Parameters
        ----------
        filename : str
            File path to read in as a table.

        delimiter : str
            Character used to delineate columns.

        format : str
            The format of the input file.

        Returns
        -------
        AssociationPool
            The ``AssociationPool`` representation of the file.
        """
        table = super().read(
            filename,
            delimiter=delimiter,
            format=format,
            converters=convert_to_str,
            **kwargs,
        )

        # If anything has been masked, just fill
        table = table.filled("null")

        # Lowercase the column names
        # Note: Cannot do in-place because modifying the
        #       list while iterating.
        columns = [column for name, column in table.columns.items()]
        for c in columns:
            c.name = c.name.lower()

        table.meta["pool_file"] = filename
        return table

    def write(self, *args, **kwargs):
        """Write the pool to a file.

        Parameters
        ----------
        output : str, file-like
            The output file or file-like object.

        delimiter : str
            The string to use to delineate columns.
            Default is '|'.

        format : str
            The format the file should be written in.
            Default is 'ascii'.

        args, kwargs : obj
            Other parameters that ``astropy.io.ascii.write`` can accept.
        """
        delimiter = kwargs.pop("delimiter", DEFAULT_DELIMITER)
        format = kwargs.pop("format", DEFAULT_FORMAT)
        try:
            super().write(*args, delimiter=delimiter, format=format, **kwargs)
        except TypeError:
            # Most likely caused by the actual `write` called
            # does not handle `delimiter`. `jsviewer` is one
            # such format.
            # So, try again without a delimiter.
            super().write(*args, format=format, **kwargs)


class PoolRow(UserDict):
    """A row from an AssociationPool

    Class to create a copy of an AssociationPool row without copying
    all of the astropy.Table.Row private attributes.
    """

    def __init__(self, init=None):
        dict_init = dict(init)
        super().__init__(dict_init)
        try:
            self.meta = init.meta
        except AttributeError:
            self.meta = dict()


def _convert_to_str():
    func, type_ = convert_numpy(str)

    def convert_func(vals):
        """Lowercase the conversion"""
        results = func(vals)
        results = [result.lower() for result in results]
        return results

    return [(convert_func, type_)]


convert_to_str = {"*": _convert_to_str()}



import asdf
import pytest
from astropy.time import Time
from roman_datamodels.datamodels import FlatRefModel, ImageModel
from roman_datamodels.maker_utils import mk_level2_image
from stpipe import crds_client

import romancal
from romancal.flatfield import FlatFieldStep
from romancal.stpipe import RomanPipeline, RomanStep


@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_open_model(step_class, tmp_path):
    """
    Test that the class is properly hooked up to datamodels.open.
    More comprehensive tests can be found in romancal.datamodels.tests,
    this is just a smoke test of the integration.
    """
    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        imod = mk_level2_image(shape=(20, 20))
        af.tree = {"roman": imod}
        af.write_to(file_path)

    step = step_class()
    with step.open_model(file_path) as model:
        assert model.meta.telescope == "ROMAN"


@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_get_reference_file(step_class):
    """
    Test that CRDS is properly integrated.
    """
    im = mk_level2_image(shape=(20, 20))
    # This will be brittle while we're using the dev server.
    # If this test starts failing mysteriously, check the
    # metadata values against the flat rmap.
    im.meta.instrument.optical_element = "F158"
    im.meta.exposure.start_time = Time("2021-01-01T12:00:00")
    model = ImageModel(im)

    step = step_class()
    reference_path = step.get_reference_file(model, "flat")

    with step.open_model(reference_path) as reference_model:
        assert isinstance(reference_model, FlatRefModel)


@pytest.mark.skip(reason="There are no grism flats.")
@pytest.mark.parametrize("step_class", [RomanPipeline, RomanStep])
def test_get_reference_file_spectral(step_class):
    """
    Test that CRDS is properly integrated.
    """
    im = mk_level2_image(shape=(20, 20))
    # This will be brittle while we're using the dev server.
    # If this test starts failing mysteriously, check the
    # metadata values against the flat rmap.
    im.meta.instrument.optical_element = "GRISM"
    im.meta.exposure.start_time = Time("2021-01-01T12:00:00")
    model = ImageModel(im)

    step = step_class()
    reference_path = step.get_reference_file(model, "flat")

    with step.open_model(reference_path) as reference_model:
        assert isinstance(reference_model, FlatRefModel)
        assert reference_model.meta.instrument.optical_element == "GRISM"


def test_log_messages(tmp_path):
    class LoggingStep(RomanStep):
        def process(self):
            self.log.warning("Splines failed to reticulate")
            return ImageModel(mk_level2_image(shape=(20, 20)))

    result = LoggingStep().run()
    assert any("Splines failed to reticulate" in l for l in result.cal_logs)


def test_crds_meta():
    """Test that context and software versions are set"""

    im = ImageModel(mk_level2_image(shape=(20, 20)))
    result = FlatFieldStep.call(im)

    assert result.meta.ref_file.crds.sw_version == crds_client.get_svn_version()
    assert result.meta.ref_file.crds.context_used == crds_client.get_context_used(
        result.crds_observatory
    )


def test_calibration_software_version():
    """Test that calibration_software_version is updated when a step is run"""

    class NullStep(RomanStep):
        def process(self, input):
            return input

    im = ImageModel(mk_level2_image(shape=(20, 20)))
    im.meta.calibration_software_version = "junkversion"

    result = NullStep.call(im)

    assert result.meta.calibration_software_version == romancal.__version__



"""
Module for parsing APCORR and ABVEGAOFFSET reference file data.
"""

import logging

import numpy as np
from astropy.utils import lazyproperty
from roman_datamodels.datamodels import ImageModel, MosaicModel

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class ReferenceData:
    """
    Class for reference file data needed by `SourceCatalogStep`.

    .. note::
        This class currently uses placeholder values as reference file
        data is not yet available.

    Parameters
    ----------
    model : `ImageModel` or `MosaicModel`
        A level-2 or level-3 image data model.

    aperture_ee : tuple of 3 int
        The aperture encircled energies to be used for aperture
        photometry. The values must be 3 strictly-increasing integers.

        Valid values will be defined in the APCORR reference files
        (e.g., 20, 30, 40, 50, 60, 70, or 80).

    Attributes
    ----------
    aperture_params : `dict`
        A dictionary containing the aperture parameters (radii, aperture
        corrections, and background annulus inner and outer radii).
    """

    def __init__(self, model, aperture_ee):
        if not isinstance(model, (ImageModel, MosaicModel)):
            raise ValueError("The input model must be a MosaicModel.")
        self.model = model

        self.aperture_ee = self._validate_aperture_ee(aperture_ee)

        try:
            # Level-2 data model
            self.instrument = self.model.meta.instrument.name
            self.detector = self.model.meta.instrument.detector
            self.optical_element = self.model.meta.instrument.optical_element
        except AttributeError:
            # Level-3 data model
            self.instrument = self.model.meta.basic.instrument
            self.detector = "N/A"
            self.optical_element = self.model.meta.basic.optical_element

        log.info(f"Instrument: {self.instrument}")
        log.info(f"Detector: {self.detector}")
        log.info(f"Optical Element: {self.optical_element}")

    @staticmethod
    def _validate_aperture_ee(aperture_ee):
        """
        Validate the input ``aperture_ee``.
        """
        aperture_ee = np.array(aperture_ee).astype(int)
        if not np.all(aperture_ee[1:] > aperture_ee[:-1]):
            raise ValueError("aperture_ee values must be strictly increasing")
        if len(aperture_ee) != 3:
            raise ValueError("aperture_ee must contain only 3 values")
        if np.any(np.logical_or(aperture_ee <= 0, aperture_ee >= 100)):
            raise ValueError("aperture_ee values must be between 0 and 100")
        return aperture_ee

    @lazyproperty
    def aperture_params(self):
        """
        A dictionary containing the aperture parameters (radii, aperture
        corrections, and background annulus inner and outer radii) for
        performing aperture photometry.
        """
        log.warning(
            "Reference file data mapping aperture encircled energies to "
            "radii (in pixels), background annuli sizes, and aperture "
            "corrections are not yet available. Using placeholder values."
        )

        return {
            "aperture_radii": np.array((1.0, 2.0, 3.0)),
            "aperture_corrections": np.array((1.0, 1.0, 1.0)),
            "aperture_ee": self.aperture_ee,
            "bkg_aperture_inner_radius": 5.0,
            "bkg_aperture_outer_radius": 10.0,
        }



"""
Assign a gWCS object to a science image.

"""

import logging

import gwcs.coordinate_frames as cf
from astropy import coordinates as coord
from astropy import units as u
from gwcs.wcs import WCS, Step
from roman_datamodels import datamodels as rdm

from ..stpipe import RomanStep
from . import pointing
from .utils import add_s_region, wcs_bbox_from_shape

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


__all__ = ["AssignWcsStep", "load_wcs"]


class AssignWcsStep(RomanStep):
    """Assign a gWCS object to a science image."""

    reference_file_types = ["distortion"]

    def process(self, input):
        reference_file_names = {}
        with rdm.open(input, lazy_load=False) as input_model:
            for reftype in self.reference_file_types:
                log.info(f"reftype, {reftype}")
                reffile = self.get_reference_file(input_model, reftype)
                # Check for a valid reference file
                if reffile == "N/A":
                    self.log.warning("No DISTORTION reference file found")
                    self.log.warning("Assign WCS step will be skipped")
                    result = input_model.copy()
                    result.meta.cal_step.assign_wcs = "SKIPPED"
                    return result

                reference_file_names[reftype] = reffile if reffile else ""
            log.info("Using reference files: %s for assign_wcs", reference_file_names)
            result = load_wcs(input_model, reference_file_names)

        if self.save_results:
            try:
                self.suffix = "assignwcs"
            except AttributeError:
                self["suffix"] = "assignwcs"

        return result


def load_wcs(input_model, reference_files=None):
    """Create a gWCS object and store it in ``Model.meta``.

    Parameters
    ----------
    input_model : `~roman_datamodels.datamodels.WfiImage`
        The exposure.
    reference_files : dict
        A dict {reftype: reference_file_name} containing all
        reference files that apply to this exposure.

    Returns
    -------
    output_model : `~roman_datamodels.ImageModel`
        The input image file with attached gWCS object.
        The input_model is modified in place.
    """
    output_model = input_model

    if reference_files is not None:
        for ref_type, ref_file in reference_files.items():
            reference_files[ref_type] = (
                ref_file if ref_file not in ["N/A", ""] else None
            )
    else:
        reference_files = {}

    # Frames
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    # Transforms between frames
    distortion = wfi_distortion(output_model, reference_files)
    tel2sky = pointing.v23tosky(output_model)

    pipeline = [
        Step(detector, distortion),
        Step(v2v3, tel2sky),
        Step(world, None),
    ]
    wcs = WCS(pipeline)
    if wcs.bounding_box is None:
        wcs.bounding_box = wcs_bbox_from_shape(output_model.data.shape)

    output_model.meta["wcs"] = wcs

    # update S_REGION
    add_s_region(output_model)

    output_model.meta.cal_step["assign_wcs"] = "COMPLETE"

    return output_model


def wfi_distortion(model, reference_files):
    """
    Create the "detector" to "v2v3" transform for WFI

    Parameters
    ----------
    model : `~roman_datamodels.datamodels.WfiImage`
        The data model for processing
    reference_files : dict
        A dict {reftype: reference_file_name} containing all
        reference files that apply to this exposure.

    Returns
    -------
    The transform model
    """

    dist = rdm.DistortionRefModel(reference_files["distortion"])
    transform = dist.coordinate_distortion_transform

    try:
        bbox = transform.bounding_box
    except NotImplementedError:
        # Check if the transform in the reference file has a ``bounding_box``.
        # If not set a ``bounding_box`` equal to the size of the image after
        # assembling all distortion corrections.
        bbox = None
    dist.close()

    if bbox is None:
        transform.bounding_box = wcs_bbox_from_shape(model.data.shape)
    else:
        transform.bounding_box = bbox

    return transform



""" Utilities for product manipulation."""

import copy
import logging
import warnings
from collections import Counter, defaultdict

from .. import config
from .diff import compare_product_membership

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def sort_by_candidate(asns):
    """Sort associations by candidate

    Parameters
    ----------
    asns : [Association[,...]]
        List of associations

    Returns
    -------
    sorted_by_candidate : [Associations[,...]]
        New list of the associations sorted.

    Notes
    -----
    The current definition of candidates allows strictly lexigraphical
    sorting:
    aXXXX > cXXXX > oXXX

    If this changes, a comparison function will need be implemented
    """
    return sorted(asns, key=lambda asn: asn["asn_id"])


def get_product_names(asns):
    """Return product names from associations and flag duplicates

    Parameters
    ----------
    asns : [`Association`[, ...]]

    Returns
    -------
    product_names, duplicates : set(str[, ...]), [str[,...]]
        2-tuple consisting of the set of product names and the list of duplicates.
    """
    product_names = [asn["products"][0]["name"] for asn in asns]

    dups = [name for name, count in Counter(product_names).items() if count > 1]
    if dups:
        logger.debug("Duplicate product names: %s", dups)

    return set(product_names), dups


def prune_duplicate_associations(asns):
    """Remove duplicate associations in favor of lower level versions

    Main use case: For Level 3 associations, multiple associations with the
    same membership, but different levels, can be created. Remove duplicate
    associations of higher level.

    The assumption is that there is only one product per association, before
    merging.

    Parameters
    ----------
    asns : [Association[,...]]
        Associations to prune

    Returns
    -------
    pruned : [Association[,...]]
        Pruned list of associations

    """
    ordered_asns = sort_by_candidate(asns)
    pruned = list()
    while True:
        try:
            original = ordered_asns.pop()
        except IndexError:
            break
        pruned.append(original)
        to_prune = list()
        for asn in ordered_asns:
            try:
                compare_product_membership(original["products"][0], asn["products"][0])
            except AssertionError:
                continue
            to_prune.append(asn)
        for prune in to_prune:
            ordered_asns.remove(prune)

    return pruned


def prune_duplicate_products(asns):
    """Remove duplicate products in favor of higher level versions

    The assumption is that there is only one product per association, before
    merging

    Parameters
    ----------
    asns: [Association[,...]]
        Associations to prune

    Returns
    pruned: [Association[,...]]
        Pruned list of associations

    """
    product_names, dups = get_product_names(asns)
    if not dups:
        return asns

    warnings.warn(f"Duplicate associations exist: {dups}", RuntimeWarning)
    if config.DEBUG:
        warnings.warn(
            'Duplicate associations will have "dupXXX" prepended to their names, where'
            ' "XXX" is a 3-digit sequence.'
        )
    else:
        warnings.warn(
            "Duplicates will be removed, leaving only one of each.", RuntimeWarning
        )

    pruned = copy.copy(asns)
    to_prune = defaultdict(list)
    for asn in asns:
        product_name = asn["products"][0]["name"]
        if product_name in dups:
            to_prune[product_name].append(asn)

    dup_count = 0
    for product_name, asns_to_prune in to_prune.items():
        asns_to_prune = sort_by_candidate(asns_to_prune)
        for asn in asns_to_prune[1:]:
            if config.DEBUG:
                dup_count += 1
                asn.asn_name = f"dup{dup_count:03d}_{asn.asn_name}"
            else:
                pruned.remove(asn)

    return pruned



import pytest


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "webbpsf: mark test as requiring webbpsf data to run"
    )


def pytest_addoption(parser):
    parser.addoption(
        "--webbpsf", action="store_true", default=False, help="run webbpsf tests"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--webbpsf"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_webbpsf = pytest.mark.skip(reason="need --webbpsf option to run")
    for item in items:
        if "webbpsf" in item.keywords:
            item.add_marker(skip_webbpsf)



import logging
import warnings

from astropy import units as u

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def photom_io(input_model, photom_metadata):
    """
    Combine photometric scalar conversion factors and add to the science metadata.

    Parameters
    ----------
    input_model : Roman level 2 image datamodel
        Input Roman datamodel

    photom_metadata : dict
        Set of matched photom meta data keys

    Returns
    -------

    """
    # Get the scalar conversion factor.
    conversion = photom_metadata["photmjsr"]  # unit is MJy / sr

    # Store the conversion factor in the meta data
    log.info(f"photmjsr value: {conversion:.6g}")
    input_model.meta.photometry.conversion_megajanskys = conversion
    input_model.meta.photometry.conversion_microjanskys = conversion.to(
        u.microjansky / u.arcsecond**2
    )

    # Get the scalar conversion uncertainty factor
    uncertainty_conv = photom_metadata["uncertainty"]

    # Store the uncertainty conversion factor in the meta data
    log.info(f"uncertainty value: {uncertainty_conv:.6g}")
    input_model.meta.photometry.conversion_megajanskys_uncertainty = uncertainty_conv
    input_model.meta.photometry.conversion_microjanskys_uncertainty = (
        uncertainty_conv.to(u.microjansky / u.arcsecond**2)
    )

    # Return updated input model
    return input_model


def save_area_info(input_model, photom_parameters):
    """
    Read the pixel area value in the photom parameters, then convert and
    copy them to the metadata of the input datamodel.

    Parameters
    ----------
    input_model : Roman level 2 image datamodel
        Input Roman datamodel

    photom_parameters : dict
        Photom parameter object
    """

    # Load the average pixel area values from the photom reference file header
    area_ster = photom_parameters["pixelareasr"]
    area_a2 = photom_parameters["pixelareasr"].to(u.arcsecond**2)

    # Copy the pixel area values to the input model
    log.debug(f"pixelarea_steradians = {area_ster}, pixelarea_arcsecsq = {area_a2}")
    input_model.meta.photometry.pixelarea_arcsecsq = area_a2
    input_model.meta.photometry.pixelarea_steradians = area_ster

    # Return updated input model
    return input_model


def apply_photom(input_model, photom):
    """
    Retrieve the conversion factor from the photom reference datamodel
    that is appropriate to the instrument mode. The scalar factor is written
    to the photmjsr and uncertainty keywords in the model.

    For WFI, matching is based on optical_element.

    Parameters
    ----------
    input_model : Roman level 2 image datamodel
        Input Roman datamodel

    photom : Roman Photom reference datamodel
        Photom Roman datamodel

    Returns
    -------
    output_model : Roman level 2 image datamodel
        Output Roman datamodel with the flux calibration keywords updated
    """
    # Obtain photom parameters for the selected optical element
    try:
        photom_parameters = photom.phot_table[
            input_model.meta.instrument.optical_element.upper()
        ]
    except KeyError:
        warnings.warn(
            "No matching photom parameters for"
            f" {input_model.meta.instrument.optical_element}"
        )
        return input_model

    # Copy pixel area information to output datamodel
    output_model = save_area_info(input_model, photom_parameters)

    # Copy conversions to output model
    output_model = photom_io(output_model, photom_parameters)

    # Return updated output model
    return output_model



import os
from pathlib import Path

import numpy as np
import pytest
from astropy.units import Quantity

from romancal.datamodels import ModelContainer
from romancal.outlier_detection import OutlierDetectionStep, outlier_detection


@pytest.fixture()
def clean_up_after_test():
    """
    Clean up working directory after test.
    """

    def _clean_up_working_directory(pattern="*.asdf"):
        """
        Clean up the working directory by removing files matching the specified pattern.
        Parameters
        ----------
        pattern : str, optional
            The pattern of the file pattern to match (default is "*.asdf")

        Returns
        -------
        None

        """
        for x in (Path(os.getcwd())).glob(pattern):
            if x.exists():
                x.unlink()

    return _clean_up_working_directory


@pytest.mark.parametrize(
    "input_models",
    [
        list(),
        "",
        None,
    ],
)
def test_outlier_raises_error_on_invalid_input_models(input_models, caplog):
    """Test that OutlierDetection logs out a WARNING if input is invalid."""

    OutlierDetectionStep.call(input_models)

    assert "WARNING" in [x.levelname for x in caplog.records]


def test_outlier_skips_step_on_invalid_number_of_elements_in_input(base_image):
    """Test that OutlierDetection skips processing when provided with an invalid number of elements in the input,
    and sets the appropriate metadata for the skipped step."""
    img = base_image()

    res = OutlierDetectionStep.call(ModelContainer([img]))

    assert all(x.meta.cal_step.outlier_detection == "SKIPPED" for x in res)


def test_outlier_skips_step_on_exposure_type_different_from_wfi_image(base_image):
    """
    Test if the outlier detection step is skipped when the exposure type is different from WFI image.
    """
    img_1 = base_image()
    img_1.meta.exposure.type = "WFI_PRISM"
    img_2 = base_image()
    img_2.meta.exposure.type = "WFI_PRISM"

    res = OutlierDetectionStep.call(ModelContainer([img_1, img_2]))

    assert all(x.meta.cal_step.outlier_detection == "SKIPPED" for x in res)


def test_outlier_valid_input_asn(tmp_path, base_image, create_mock_asn_file):
    """
    Test that OutlierDetection runs with valid ASN file as input.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_1.save(tmp_path / "img_1.asdf")
    img_2 = base_image()
    img_1.meta.filename = "img_2.asdf"
    img_2.save(tmp_path / "img_2.asdf")

    asn_filepath = create_mock_asn_file(tmp_path)

    res = OutlierDetectionStep.call(
        asn_filepath,
        output_dir=tmp_path.as_posix(),
        in_memory=True,
        resample_data=False,
    )

    # assert step.skip is False
    assert all(x.meta.cal_step.outlier_detection == "COMPLETE" for x in res)


def test_outlier_valid_input_modelcontainer(tmp_path, base_image):
    """
    Test that OutlierDetection runs with valid ModelContainer as input.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"

    mc = ModelContainer([img_1, img_2])

    res = OutlierDetectionStep.call(
        mc,
        in_memory=True,
        resample_data=False,
    )

    assert all(x.meta.cal_step.outlier_detection == "COMPLETE" for x in res)


@pytest.mark.parametrize(
    "pars",
    [
        {
            "weight_type": "exptime",
            "pixfrac": 1.0,
            "kernel": "square",
            "fillval": "INDEF",
            "nlow": 0,
            "nhigh": 0,
            "maskpt": 0.7,
            "grow": 1,
            "snr": "4.0 3.0",
            "scale": "0.5 0.4",
            "backg": 0.0,
            "kernel_size": "7 7",
            "save_intermediate_results": False,
            "resample_data": True,
            "good_bits": 0,
            "allowed_memory": None,
            "in_memory": True,
            "make_output_path": None,
            "resample_suffix": "i2d",
        },
        {
            "weight_type": "exptime",
            "save_intermediate_results": True,
            "make_output_path": None,
            "resample_suffix": "some_other_suffix",
        },
    ],
)
def test_outlier_init_default_parameters(pars, base_image):
    """
    Test parameter setting on initialization for OutlierDetection.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    input_models = ModelContainer([img_1])

    step = outlier_detection.OutlierDetection(input_models, **pars)

    assert step.input_models == input_models
    assert step.outlierpars == pars
    assert step.make_output_path == pars["make_output_path"]
    assert step.resample_suffix == f"_outlier_{pars['resample_suffix']}.asdf"


def test_outlier_do_detection_write_files_to_custom_location(
    tmp_path, base_image, clean_up_after_test
):
    """
    Test that OutlierDetection can create files on disk in a custom location.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"
    input_models = ModelContainer([img_1, img_2])

    outlier_step = OutlierDetectionStep()
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path.as_posix()
    # make sure files are written out to disk
    outlier_step.in_memory = False

    pars = {
        "weight_type": "exptime",
        "pixfrac": 1.0,
        "kernel": "square",
        "fillval": "INDEF",
        "nlow": 0,
        "nhigh": 0,
        "maskpt": 0.7,
        "grow": 1,
        "snr": "4.0 3.0",
        "scale": "0.5 0.4",
        "backg": 0.0,
        "kernel_size": "7 7",
        "save_intermediate_results": False,
        "resample_data": False,
        "good_bits": 0,
        "allowed_memory": None,
        "in_memory": outlier_step.in_memory,
        "make_output_path": outlier_step.make_output_path,
        "resample_suffix": "i2d",
    }

    # meta.filename for the median image created by OutlierDetection.do_detection()
    median_path = tmp_path / "drizzled_median.asdf"

    outlier_files_path = [
        median_path,
    ]

    step = outlier_detection.OutlierDetection(input_models, **pars)
    step.do_detection()

    assert all(x.exists() for x in outlier_files_path)

    clean_up_after_test("*.asdf")


def test_outlier_do_detection_find_outliers(tmp_path, base_image, clean_up_after_test):
    """
    Test that OutlierDetection can find outliers.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"

    # add outliers
    img_1_input_coords = np.array(
        [(5, 45), (25, 25), (45, 85), (65, 65), (85, 5)], dtype=[("x", int), ("y", int)]
    )
    img_2_input_coords = np.array(
        [(15, 25), (35, 5), (75, 65), (95, 45), (99, 5)], dtype=[("x", int), ("y", int)]
    )
    img_1.data[img_1_input_coords["x"], img_1_input_coords["y"]] = Quantity(
        100000, "DN / s"
    )
    img_2.data[img_2_input_coords["x"], img_2_input_coords["y"]] = Quantity(
        100000, "DN / s"
    )

    input_models = ModelContainer([img_1, img_2])

    outlier_step = OutlierDetectionStep()
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path.as_posix()
    # make sure files are written out to disk
    outlier_step.in_memory = False

    pars = {
        "weight_type": "exptime",
        "pixfrac": 1.0,
        "kernel": "square",
        "fillval": "INDEF",
        "nlow": 0,
        "nhigh": 0,
        "maskpt": 0.7,
        "grow": 1,
        "snr": "4.0 3.0",
        "scale": "0.5 0.4",
        "backg": 0.0,
        "kernel_size": "7 7",
        "save_intermediate_results": False,
        "resample_data": False,
        "good_bits": 0,
        "allowed_memory": None,
        "in_memory": outlier_step.in_memory,
        "make_output_path": outlier_step.make_output_path,
        "resample_suffix": "i2d",
    }

    detection_step = outlier_detection.OutlierDetection
    step = detection_step(input_models, **pars)

    step.do_detection()

    # get flagged outliers coordinates from DQ array
    img_1_outlier_output_coords = np.where(step.input_models[0].dq > 0)

    # reformat output and input coordinates and sort by x coordinate
    outliers_output_coords = np.array(
        list(zip(*img_1_outlier_output_coords)), dtype=[("x", int), ("y", int)]
    )
    outliers_input_coords = np.concatenate((img_1_input_coords, img_2_input_coords))

    outliers_output_coords.sort(axis=0)
    outliers_input_coords.sort(axis=0)

    # assert all(outliers_input_coords == outliers_output_coords) doesn't work with python 3.9
    assert all(o == i for i, o in zip(outliers_input_coords, outliers_output_coords))

    clean_up_after_test("*.asdf")


def test_outlier_do_detection_do_not_find_outliers_in_identical_images(
    tmp_path, base_image, clean_up_after_test, caplog
):
    """
    Test that OutlierDetection does not flag any outliers in the DQ array if images are identical.
    """
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    # add outliers
    img_1_input_coords = np.array(
        [(5, 45), (25, 25), (45, 85), (65, 65), (85, 5)], dtype=[("x", int), ("y", int)]
    )
    img_1.data[img_1_input_coords["x"], img_1_input_coords["y"]] = Quantity(
        100000, "DN / s"
    )

    img_2 = img_1.copy()
    img_2.meta.filename = "img_2.asdf"
    img_3 = img_1.copy()
    img_3.meta.filename = "img_3.asdf"

    input_models = ModelContainer([img_1, img_2, img_3])

    outlier_step = OutlierDetectionStep()
    # set output dir for all files created by the step
    outlier_step.output_dir = tmp_path.as_posix()
    # make sure files are written out to disk
    outlier_step.in_memory = False

    pars = {
        "weight_type": "exptime",
        "pixfrac": 1.0,
        "kernel": "square",
        "fillval": "INDEF",
        "nlow": 0,
        "nhigh": 0,
        "maskpt": 0.7,
        "grow": 1,
        "snr": "4.0 3.0",
        "scale": "0.5 0.4",
        "backg": 0.0,
        "kernel_size": "7 7",
        "save_intermediate_results": False,
        "resample_data": False,
        "good_bits": 0,
        "allowed_memory": None,
        "in_memory": outlier_step.in_memory,
        "make_output_path": outlier_step.make_output_path,
        "resample_suffix": "i2d",
    }

    detection_step = outlier_detection.OutlierDetection
    step = detection_step(input_models, **pars)

    step.do_detection()

    # assert that log shows no new outliers detected
    assert "New pixels flagged as outliers: 0 (0.00%)" in {
        x.message for x in caplog.records
    }
    # assert that DQ array has nothing flagged as outliers
    assert [np.count_nonzero(x.dq) for x in step.input_models] == [0, 0, 0]


@pytest.mark.parametrize(
    "input_type",
    [
        "ModelContainer",
        "ASNFile",
        "DataModelList",
        "ASDFFilenameList",
    ],
)
def test_skymatch_always_returns_modelcontainer_with_updated_datamodels(
    input_type,
    base_image,
    tmp_path,
    create_mock_asn_file,
):
    """Test that the OutlierDetectionStep always returns a ModelContainer
    with updated data models after processing different input types."""

    os.chdir(tmp_path)
    img_1 = base_image()
    img_1.meta.filename = "img_1.asdf"
    img_2 = base_image()
    img_2.meta.filename = "img_2.asdf"

    mc = ModelContainer([img_1, img_2])

    mc.save(dir_path=tmp_path)

    step_input_map = {
        "ModelContainer": mc,
        "ASNFile": create_mock_asn_file(
            tmp_path,
            members_mapping=[
                {"expname": img_1.meta.filename, "exptype": "science"},
                {"expname": img_2.meta.filename, "exptype": "science"},
            ],
        ),
        "DataModelList": [img_1, img_2],
        "ASDFFilenameList": [img_1.meta.filename, img_2.meta.filename],
    }

    step_input = step_input_map.get(input_type)

    res = OutlierDetectionStep.call(step_input)

    assert isinstance(res, ModelContainer)
    assert all(x.meta.cal_step.outlier_detection == "COMPLETE" for x in res)



import warnings

import numpy as np
import pytest
from astropy import units as u
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import ImageModel, WfiImgPhotomRefModel

from romancal.photom import PhotomStep, photom


def create_photom_wfi_image(min_r=3.1, delta=0.1):
    """Create a photom table for WFI.

    Parameters
    ----------
    min_r : float
        Minimum value to assign when populating the photometry array.

    max_r : float
        Maximum value to assign when populating the photometry array.

    Returns
    -------
    photom_model : Roman Photom reference datamodel
        photom roman datamodel

    """

    # Keyword and eleents list for phot_table construction
    optical_element = [
        "F062",
        "F087",
        "F106",
        "F129",
        "F146",
        "F158",
        "F184",
        "F213",
        "GRISM",
        "PRISM",
        "DARK",
    ]
    none_type_elements = ["GRISM", "PRISM", "DARK"]
    keyword = ["photmjsr", "uncertainty", "pixelareasr"]
    nrows = len(optical_element)

    # Create sample photometry keyword values
    photmjsr = (
        np.linspace(min_r, min_r + (nrows - 1.0) * delta, nrows)
        * u.megajansky
        / u.steradian
    )
    uncertainty = (
        np.linspace(min_r / 20.0, min_r / 20.0 + (nrows - 1.0) * delta / 20.0, nrows)
        * u.megajansky
        / u.steradian
    )

    # Create sample area keyword values
    area_ster = 2.31307642258977e-14 * u.steradian
    pixelareasr = np.ones(nrows, dtype=np.float64) * area_ster

    # Bundle values into a list
    values = list(zip(photmjsr, uncertainty, pixelareasr))

    # Create dictionary containing all values
    reftab = {}
    for element_idx, element in enumerate(optical_element):
        key_dict = {}
        for key_idx, key in enumerate(keyword):
            key_dict[key] = values[element_idx][key_idx]
        # GRISM, PRISM, and DARK optical elements shuld have their
        # photomeetry keywords set to None
        if element in none_type_elements:
            key_dict["photmjsr"] = None
            key_dict["uncertainty"] = None
        reftab[element] = key_dict

    # Create default datamodel
    photom_model = maker_utils.mk_wfi_img_photom()

    # Copy values above into defautl datamodel
    photom_model.phot_table = reftab

    return photom_model


def test_no_photom_match():
    """Test apply_photom warning for no match"""

    # Create sample WFI Level 2 science datamodel
    input_model = maker_utils.mk_level2_image()

    # Create photom reference datamodel
    photom_model = create_photom_wfi_image(min_r=3.1, delta=0.1)

    # Remove key for failed test (that won't fail validation)
    photom_model.phot_table.pop("F146")

    # Select optical element
    input_model.meta.instrument.optical_element = "F146"

    # Set bad values which would be overwritten by apply_photom
    input_model.meta.photometry.pixelarea_steradians = -1.0 * u.sr
    input_model.meta.photometry.conversion_megajanskys = (
        -1.0 * u.megajansky / u.steradian
    )
    input_model.meta.photometry.conversion_microjanskys_uncertainty = (
        -1.0 * u.microjansky / u.arcsecond**2
    )

    with warnings.catch_warnings(record=True) as caught:
        # Look for now non existent F146 optical element
        output_model = photom.apply_photom(input_model, photom_model)

        # Assert warning key matches that of the input file
        assert (
            str(caught[0].message).split()[-1]
            == input_model.meta.instrument.optical_element
        )

        # Assert that photom elements are not updated
        assert output_model.meta.photometry.pixelarea_steradians == -1.0 * u.sr
        assert (
            output_model.meta.photometry.conversion_megajanskys
            == -1.0 * u.megajansky / u.steradian
        )
        assert (
            output_model.meta.photometry.conversion_microjanskys_uncertainty
            == -1.0 * u.microjansky / u.arcsecond**2
        )


def test_apply_photom1():
    """Test apply_photom applies correct metadata"""

    # Create sample WFI Level 2 science datamodel
    input_model = maker_utils.mk_level2_image()

    # Create photom reference datamodel
    photom_model = create_photom_wfi_image(min_r=3.1, delta=0.1)

    # Select optical element
    input_model.meta.instrument.optical_element = "F146"

    # Apply photom correction for optical element F146
    output_model = photom.apply_photom(input_model, photom_model)

    # Set reference photometry
    area_ster = 2.31307642258977e-14 * u.steradian
    area_a2 = 0.000984102303070964 * u.arcsecond * u.arcsecond

    # Tests for pixel areas
    assert np.isclose(
        output_model.meta.photometry.pixelarea_steradians.value,
        area_ster.value,
        atol=1.0e-7,
    )
    assert output_model.meta.photometry.pixelarea_steradians.unit == area_ster.unit
    assert np.isclose(
        output_model.meta.photometry.pixelarea_arcsecsq.value,
        area_a2.value,
        atol=1.0e-7,
    )
    assert output_model.meta.photometry.pixelarea_arcsecsq.unit == area_a2.unit

    # Set reference photometry
    phot_ster = 3.5 * u.megajansky / u.steradian
    phot_a2 = phot_ster.to(u.microjansky / u.arcsecond**2)

    # Tests for photometry
    assert np.isclose(
        output_model.meta.photometry.conversion_megajanskys.value,
        phot_ster.value,
        atol=1.0e-7,
    )
    assert output_model.meta.photometry.conversion_megajanskys.unit == phot_ster.unit
    assert np.isclose(
        output_model.meta.photometry.conversion_microjanskys.value,
        phot_a2.value,
        atol=1.0e-7,
    )
    assert output_model.meta.photometry.conversion_microjanskys.unit == phot_a2.unit

    # Set reference photometric uncertainty
    muphot_ster = 0.175 * u.megajansky / u.steradian
    muphot_a2 = muphot_ster.to(u.microjansky / u.arcsecond**2)

    # Tests for photometric uncertainty
    assert np.isclose(
        output_model.meta.photometry.conversion_megajanskys_uncertainty.value,
        muphot_ster.value,
        atol=1.0e-7,
    )
    assert (
        output_model.meta.photometry.conversion_megajanskys_uncertainty.unit
        == muphot_ster.unit
    )
    assert np.isclose(
        output_model.meta.photometry.conversion_microjanskys_uncertainty.value,
        muphot_a2.value,
        atol=1.0e-7,
    )
    assert (
        output_model.meta.photometry.conversion_microjanskys_uncertainty.unit
        == muphot_a2.unit
    )


def test_apply_photom2():
    """Test apply_photom does not change data values"""

    # Create sample WFI Level 2 science datamodel
    input_model = maker_utils.mk_level2_image()

    # Create photom reference datamodel
    photom_model = create_photom_wfi_image(min_r=3.1, delta=0.1)

    # Apply photom correction
    output_model = photom.apply_photom(input_model, photom_model)

    # Select pixel for comparison
    shape = input_model.data.shape
    ix = shape[1] // 2
    iy = shape[0] // 2

    # Test that the data has not changed
    assert np.allclose(output_model.data[iy, ix], input_model.data[iy, ix], rtol=1.0e-7)


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_photom_step_interface(instrument, exptype):
    """Test that the basic inferface works for data requiring a photom reffile"""

    # Create a small area for the file
    shape = (20, 20)

    # Create input model
    wfi_image = maker_utils.mk_level2_image(shape=shape)
    wfi_image_model = ImageModel(wfi_image)

    # Create photom model
    photom = maker_utils.mk_wfi_img_photom()
    photom_model = WfiImgPhotomRefModel(photom)

    # Run photom correction step
    result = PhotomStep.call(wfi_image_model, override_photom=photom_model)

    assert (result.data == wfi_image.data).all()
    assert result.data.shape == shape
    if exptype == "WFI_IMAGE":
        assert result.meta.cal_step.photom == "COMPLETE"
    else:
        assert result.meta.cal_step.photom == "SKIPPED"

    # Run photom correction step with reffile as N/A
    result = PhotomStep.call(wfi_image_model, override_photom="N/A")

    assert result.meta.cal_step.photom == "SKIPPED"


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_PRISM"),
    ],
)
def test_photom_step_interface_spectroscopic(instrument, exptype):
    """
    Test apply_photom properly populates photometric keywords for spectroscopic data
    """

    # Create a small area for the file
    shape = (20, 20)

    # Create input node
    wfi_image = maker_utils.mk_level2_image(shape=shape)

    # Select exposure type and optical element
    wfi_image.meta.exposure.type = "WFI_PRISM"
    wfi_image.meta.instrument.optical_element = "PRISM"

    # Set photometric values for spectroscopic data
    wfi_image.meta.photometry.pixelarea_steradians = 2.31307642258977e-14 * u.steradian
    wfi_image.meta.photometry.pixelarea_arcsecsq = (
        0.000984102303070964 * u.arcsecond * u.arcsecond
    )
    wfi_image.meta.photometry.conversion_megajanskys = (
        -99999 * u.megajansky / u.steradian
    )
    wfi_image.meta.photometry.conversion_megajanskys_uncertainty = (
        -99999 * u.megajansky / u.steradian
    )
    wfi_image.meta.photometry.conversion_microjanskys = (
        -99999 * u.microjansky / u.arcsecond**2
    )
    wfi_image.meta.photometry.conversion_microjanskys_uncertainty = (
        -99999 * u.microjansky / u.arcsecond**2
    )

    # Create input model
    wfi_image_model = ImageModel(wfi_image)

    # Create photom model
    photom = maker_utils.mk_wfi_img_photom()
    photom_model = WfiImgPhotomRefModel(photom)

    # Run photom correction step
    result = PhotomStep.call(wfi_image_model, override_photom=photom_model)

    # Test that the data has not changed
    assert np.allclose(result.data, wfi_image_model.data, rtol=1.0e-7)

    # Test that keywords are properly overwritten
    assert result.meta.photometry.conversion_megajanskys is None
    assert result.meta.photometry.conversion_microjanskys is None
    assert result.meta.photometry.conversion_megajanskys_uncertainty is None
    assert result.meta.photometry.conversion_microjanskys_uncertainty is None
    assert result.meta.photometry.pixelarea_steradians is None
    assert result.meta.photometry.pixelarea_arcsecsq is None



"""
Roman pipeline step for image alignment.
"""

import os
import weakref
from pathlib import Path

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
from roman_datamodels import datamodels as rdm
from tweakwcs.correctors import JWSTWCSCorrector
from tweakwcs.imalign import align_wcs
from tweakwcs.matchutils import XYXYMatch

from romancal.lib.basic_utils import is_association

# LOCAL
from ..datamodels import ModelContainer
from ..stpipe import RomanStep
from . import astrometric_utils as amutils


def _oxford_or_str_join(str_list):
    nelem = len(str_list)
    if not nelem:
        return "N/A"
    str_list = list(map(repr, str_list))
    if nelem == 1:
        return str_list
    elif nelem == 2:
        return f"{str_list[0]} or {str_list[1]}"
    else:
        return ", ".join(map(repr, str_list[:-1])) + ", or " + repr(str_list[-1])


SINGLE_GROUP_REFCAT = ["GAIADR3", "GAIADR2", "GAIADR1"]
_SINGLE_GROUP_REFCAT_STR = _oxford_or_str_join(SINGLE_GROUP_REFCAT)
DEFAULT_ABS_REFCAT = SINGLE_GROUP_REFCAT[0]
ALIGN_TO_ABS_REFCAT = True

__all__ = ["TweakRegStep"]


class TweakRegStep(RomanStep):
    """
    TweakRegStep: Image alignment based on catalogs of sources detected in
    input images.
    """

    class_alias = "tweakreg"

    spec = f"""
        use_custom_catalogs = boolean(default=False) # Use custom user-provided catalogs?
        catalog_format = string(default='ascii.ecsv') # Catalog output file format
        catfile = string(default='') # Name of the file with a list of custom user-provided catalogs
        catalog_path = string(default='') # Catalog output file path
        enforce_user_order = boolean(default=False) # Align images in user specified order?
        expand_refcat = boolean(default=False) # Expand reference catalog with new sources?
        minobj = integer(default=15) # Minimum number of objects acceptable for matching
        searchrad = float(default=2.0) # The search radius in arcsec for a match
        use2dhist = boolean(default=True) # Use 2d histogram to find initial offset?
        separation = float(default=1.0) # Minimum object separation in arcsec
        tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec
        fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift') # Fitting geometry
        nclip = integer(min=0, default=3) # Number of clipping iterations in fit
        sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units
        abs_refcat = string(default='{DEFAULT_ABS_REFCAT}')  # Absolute reference
        # catalog. Options: {_SINGLE_GROUP_REFCAT_STR}
        save_abs_catalog = boolean(default=False)  # Write out used absolute astrometric reference catalog as a separate product
        abs_minobj = integer(default=15) # Minimum number of objects acceptable for matching when performing absolute astrometry
        abs_searchrad = float(default=6.0) # The search radius in arcsec for a match when performing absolute astrometry
        # We encourage setting this parameter to True. Otherwise, xoffset and yoffset will be set to zero.
        abs_use2dhist = boolean(default=True) # Use 2D histogram to find initial offset when performing absolute astrometry?
        abs_separation = float(default=0.1) # Minimum object separation in arcsec when performing absolute astrometry
        abs_tolerance = float(default=0.7) # Matching tolerance for xyxymatch in arcsec when performing absolute astrometry
        # Fitting geometry when performing absolute astrometry
        abs_fitgeometry = option('shift', 'rshift', 'rscale', 'general', default='rshift')
        abs_nclip = integer(min=0, default=3) # Number of clipping iterations in fit when performing absolute astrometry
        abs_sigma = float(min=0.0, default=3.0) # Clipping limit in sigma units when performing absolute astrometry
        output_use_model = boolean(default=True)  # When saving use `DataModel.meta.filename`
    """  # noqa: E501

    reference_file_types = []
    refcat = None

    def process(self, input):

        use_custom_catalogs = self.use_custom_catalogs

        if use_custom_catalogs:
            catdict = _parse_catfile(self.catfile)
            # if user requested the use of custom catalogs and provided a
            # valid 'catfile' file name that has no custom catalogs,
            # turn off the use of custom catalogs:
            if catdict is not None and not catdict:
                self.log.warning(
                    "'use_custom_catalogs' is set to True but 'catfile' "
                    "contains no user catalogs."
                )
                use_custom_catalogs = False

        try:
            if use_custom_catalogs and catdict:
                images = ModelContainer()
                if isinstance(input, str):
                    asn_dir = os.path.dirname(input)
                    asn_data = images.read_asn(input)
                    for member in asn_data["products"][0]["members"]:
                        filename = member["expname"]
                        member["expname"] = os.path.join(asn_dir, filename)
                        if filename in catdict:
                            member["tweakreg_catalog"] = catdict[filename]
                        elif "tweakreg_catalog" in member:
                            del member["tweakreg_catalog"]

                    images.from_asn(asn_data)
                elif is_association(input):
                    images.from_asn(input)
                else:
                    images = ModelContainer(input)
                    for im in images:
                        filename = im.meta.filename
                        if filename in catdict:
                            self.log.info(
                                f"setting "
                                f"{filename}.source_detection.tweakreg_catalog_name ="
                                f" {repr(catdict[filename])}"
                            )
                            # set catalog name only (no catalog data at this point)
                            im.meta["source_detection"] = {
                                "tweakreg_catalog_name": catdict[filename],
                            }
            else:
                images = (
                    ModelContainer([input])
                    if (
                        isinstance(input, rdm.DataModel) or str(input).endswith(".asdf")
                    )
                    else ModelContainer(input)
                )
        except TypeError as e:
            e.args = (
                "Input to tweakreg must be a list of DataModels, an "
                "association, or an already open ModelContainer "
                "containing one or more DataModels.",
            ) + e.args[1:]
            raise e

        if len(self.catalog_path) == 0:
            self.catalog_path = os.getcwd()

        self.catalog_path = Path(self.catalog_path).as_posix()
        self.log.info(f"All source catalogs will be saved to: {self.catalog_path}")

        if self.abs_refcat is None or len(self.abs_refcat.strip()) == 0:
            self.abs_refcat = DEFAULT_ABS_REFCAT

        if self.abs_refcat != DEFAULT_ABS_REFCAT:
            # Set expand_refcat to True to eliminate possibility of duplicate
            # entries when aligning to absolute astrometric reference catalog
            self.expand_refcat = True

        if len(images) == 0:
            raise ValueError("Input must contain at least one image model.")

        # Build the catalogs for input images
        for i, image_model in enumerate(images):
            if image_model.meta.exposure.type != "WFI_IMAGE":
                # Check to see if attempt to run tweakreg on non-Image data
                self.log.info("Skipping TweakReg for spectral exposure.")
                # Uncomment below once rad & input data have the cal_step tweakreg
                # image_model.meta.cal_step.tweakreg = "SKIPPED"
                return image_model

            if hasattr(image_model.meta, "source_detection"):
                is_tweakreg_catalog_present = hasattr(
                    image_model.meta.source_detection, "tweakreg_catalog"
                )
                is_tweakreg_catalog_name_present = hasattr(
                    image_model.meta.source_detection, "tweakreg_catalog_name"
                )
                if is_tweakreg_catalog_present:
                    # read catalog from structured array
                    catalog = Table(
                        np.asarray(image_model.meta.source_detection.tweakreg_catalog)
                    )
                elif is_tweakreg_catalog_name_present:
                    catalog = Table.read(
                        image_model.meta.source_detection.tweakreg_catalog_name,
                        format=self.catalog_format,
                    )
                else:
                    raise AttributeError(
                        "Attribute 'meta.source_detection.tweakreg_catalog' is missing."
                        "Please either run SourceDetectionStep or provide a"
                        "custom source catalog."
                    )
                # remove 4D numpy array from meta.source_detection
                if is_tweakreg_catalog_present:
                    del image_model.meta.source_detection["tweakreg_catalog"]
            else:
                raise AttributeError(
                    "Attribute 'meta.source_detection' is missing."
                    "Please either run SourceDetectionStep or provide a"
                    "custom source catalog."
                )

            for axis in ["x", "y"]:
                if axis not in catalog.colnames:
                    long_axis = axis + "centroid"
                    if long_axis in catalog.colnames:
                        catalog.rename_column(long_axis, axis)
                    else:
                        raise ValueError(
                            "'tweakreg' source catalogs must contain a header with "
                            "columns named either 'x' and 'y' or "
                            "'xcentroid' and 'ycentroid'."
                        )

            filename = image_model.meta.filename

            # filter out sources outside the WCS bounding box
            bb = image_model.meta.wcs.bounding_box
            x = catalog["x"]
            y = catalog["y"]
            if bb is None:
                r, d = image_model.meta.wcs(x, y)
                mask = np.isfinite(r) & np.isfinite(d)
                catalog = catalog[mask]

                n_removed_src = np.sum(np.logical_not(mask))
                if n_removed_src:
                    self.log.info(
                        f"Removed {n_removed_src} sources from {filename}'s "
                        "catalog whose image coordinates could not be "
                        "converted to world coordinates."
                    )
            else:
                # assume image coordinates of all sources within a bounding box
                # can be converted to world coordinates.
                ((xmin, xmax), (ymin, ymax)) = bb
                mask = (x > xmin) & (x < xmax) & (y > ymin) & (y < ymax)
                catalog = catalog[mask]

                n_removed_src = np.sum(np.logical_not(mask))
                if n_removed_src:
                    self.log.info(
                        f"Removed {n_removed_src} sources from {filename}'s "
                        "catalog that were outside of the bounding box."
                    )

            # set meta.tweakreg_catalog
            image_model.meta["tweakreg_catalog"] = catalog.as_array()

            nsources = len(catalog)
            if nsources == 0:
                self.log.warning(f"No sources found in {filename}.")
            else:
                self.log.info(f"Detected {len(catalog)} sources in {filename}.")

            images[i] = image_model

        # group images by their "group id":
        grp_img = list(images.models_grouped)

        self.log.info("")
        self.log.info(f"Number of image groups to be aligned: {len(grp_img):d}.")
        self.log.info("Image groups:")

        if len(grp_img) == 1 and not ALIGN_TO_ABS_REFCAT:
            self.log.info("* Images in GROUP 1:")
            for im in grp_img[0]:
                self.log.info(f"     {im.meta.filename}")
            self.log.info("")

            # we need at least two exposures to perform image alignment
            self.log.warning("At least two exposures are required for image alignment.")
            self.log.warning("Nothing to do. Skipping 'TweakRegStep'...")
            self.skip = True
            for model in images:
                model.meta.cal_step["tweakreg"] = "SKIPPED"
            return input

        elif len(grp_img) == 1 and ALIGN_TO_ABS_REFCAT:
            # create a list of WCS-Catalog-Images Info and/or their Groups:
            g = grp_img[0]
            if len(g) == 0:
                raise AssertionError("Logical error in the pipeline code.")
            group_name = _common_name(g)
            imcats = list(map(self._imodel2wcsim, g))
            # Remove the attached catalogs
            for model in g:
                model = (
                    model
                    if isinstance(model, rdm.DataModel)
                    else rdm.open(os.path.basename(model))
                )
            self.log.info(f"* Images in GROUP '{group_name}':")
            for im in imcats:
                im.meta["group_id"] = group_name
                self.log.info(f"     {im.meta['name']}")

            self.log.info("")

        elif len(grp_img) > 1:
            # create a list of WCS-Catalog-Images Info and/or their Groups:
            imcats = []
            for g in grp_img:
                if len(g) == 0:
                    raise AssertionError("Logical error in the pipeline code.")
                else:
                    group_name = _common_name(g)
                    wcsimlist = list(map(self._imodel2wcsim, g))
                    # Remove the attached catalogs
                    # for model in g:
                    #     del model.catalog
                    self.log.info(f"* Images in GROUP '{group_name}':")
                    for im in wcsimlist:
                        im.meta["group_id"] = group_name
                        # im.meta["image_model"] = group_name
                        self.log.info(f"     {im.meta['name']}")
                    imcats.extend(wcsimlist)

            self.log.info("")

            # align images:
            xyxymatch = XYXYMatch(
                searchrad=self.searchrad,
                separation=self.separation,
                use2dhist=self.use2dhist,
                tolerance=self.tolerance,
                xoffset=0,
                yoffset=0,
            )

            try:
                align_wcs(
                    imcats,
                    refcat=None or self.refcat,
                    enforce_user_order=self.enforce_user_order,
                    expand_refcat=self.expand_refcat,
                    minobj=self.minobj,
                    match=xyxymatch,
                    fitgeom=self.fitgeometry,
                    nclip=self.nclip,
                    sigma=(self.sigma, "rmse"),
                    clip_accum=True,
                )

            except ValueError as e:
                msg = e.args[0]
                if (
                    msg == "Too few input images (or groups of images) with non-empty"
                    " catalogs."
                ):
                    # we need at least two exposures to perform image alignment
                    self.log.warning(msg)
                    self.log.warning(
                        "At least two exposures are required for image alignment."
                    )
                    self.log.warning("Nothing to do. Skipping 'TweakRegStep'...")
                    for model in images:
                        model.meta.cal_step["tweakreg"] = "SKIPPED"
                    if not ALIGN_TO_ABS_REFCAT:
                        self.skip = True
                        return images
                else:
                    raise e

            except RuntimeError as e:
                msg = e.args[0]
                if msg.startswith("Number of output coordinates exceeded allocation"):
                    # we need at least two exposures to perform image alignment
                    self.log.error(msg)
                    self.log.error(
                        "Multiple sources within specified tolerance "
                        "matched to a single reference source. Try to "
                        "adjust 'tolerance' and/or 'separation' parameters."
                    )
                    self.log.warning("Skipping 'TweakRegStep'...")
                    self.skip = True
                    for model in images:
                        model.meta.cal_step.tweakreg = "SKIPPED"
                    return images
                else:
                    raise e

            for imcat in imcats:
                model = imcat.meta["image_model"]()
                if model.meta.cal_step.get("tweakreg") == "SKIPPED":
                    continue
                wcs = model.meta.wcs
                twcs = imcat.wcs
                if not self._is_wcs_correction_small(wcs, twcs):
                    # Large corrections are typically a result of source
                    # mis-matching or poorly-conditioned fit. Skip such models.
                    self.log.warning(
                        "WCS has been tweaked by more than"
                        f" {10 * self.tolerance} arcsec"
                    )

                    for model in images:
                        model.meta.cal_step["tweakreg"] = "SKIPPED"
                    if ALIGN_TO_ABS_REFCAT:
                        self.log.warning("Skipping relative alignment (stage 1)...")
                    else:
                        self.log.warning("Skipping 'TweakRegStep'...")
                        self.skip = True
                        return images

        if ALIGN_TO_ABS_REFCAT:
            # Get catalog of GAIA sources for the field
            #
            # NOTE:  If desired, the pipeline can write out the reference
            #        catalog as a separate product with a name based on
            #        whatever convention is determined by the JWST Cal Working
            #        Group.
            if self.save_abs_catalog:
                output_name = os.path.join(
                    self.catalog_path, f"fit_{self.abs_refcat.lower()}_ref.ecsv"
                )
            else:
                output_name = None

            # initial shift to be used with absolute astrometry
            self.abs_xoffset = 0
            self.abs_yoffset = 0

            self.abs_refcat = self.abs_refcat.strip()
            gaia_cat_name = self.abs_refcat.upper()

            if gaia_cat_name in SINGLE_GROUP_REFCAT:
                try:
                    ref_cat = amutils.create_astrometric_catalog(
                        images, gaia_cat_name, output=output_name
                    )
                except Exception as e:
                    self.log.warning(
                        "TweakRegStep cannot proceed because of an error that "
                        "occurred while fetching data from the VO server. "
                        f"Returned error message: '{e}'"
                    )
                    self.log.warning("Skipping 'TweakRegStep'...")
                    self.skip = True
                    for model in images:
                        model.meta.cal_step["tweakreg"] = "SKIPPED"
                    return images

            elif os.path.isfile(self.abs_refcat):
                ref_cat = Table.read(self.abs_refcat)

            else:
                raise ValueError(
                    "'abs_refcat' must be a path to an "
                    "existing file name or one of the supported "
                    f"reference catalogs: {_SINGLE_GROUP_REFCAT_STR}."
                )

            # Check that there are enough GAIA sources for a reliable/valid fit
            num_ref = len(ref_cat)
            if num_ref < self.abs_minobj:
                # Raise Exception here to avoid rest of code in this try block
                self.log.warning(
                    f"Not enough sources ({num_ref}) in the reference catalog "
                    "for the single-group alignment step to perform a fit. "
                    f"Skipping alignment to the {self.abs_refcat} reference "
                    "catalog!"
                )
            else:
                # align images:
                # Update to separation needed to prevent confusion of sources
                # from overlapping images where centering is not consistent or
                # for the possibility that errors still exist in relative overlap.
                xyxymatch_gaia = XYXYMatch(
                    searchrad=self.abs_searchrad,
                    separation=self.abs_separation,
                    use2dhist=self.abs_use2dhist,
                    tolerance=self.abs_tolerance,
                    xoffset=self.abs_xoffset,
                    yoffset=self.abs_yoffset,
                )

                # Set group_id to same value so all get fit as one observation
                # The assigned value, 987654, has been hard-coded to make it
                # easy to recognize when alignment to GAIA was being performed
                # as opposed to the group_id values used for relative alignment
                # earlier in this step.
                for imcat in imcats:
                    imcat.meta["group_id"] = 987654
                    if (
                        "fit_info" in imcat.meta
                        and "REFERENCE" in imcat.meta["fit_info"]["status"]
                    ):
                        del imcat.meta["fit_info"]

                # Perform fit
                align_wcs(
                    imcats,
                    refcat=ref_cat,
                    enforce_user_order=True,
                    expand_refcat=False,
                    minobj=self.abs_minobj,
                    match=xyxymatch_gaia,
                    fitgeom=self.abs_fitgeometry,
                    nclip=self.abs_nclip,
                    sigma=(self.abs_sigma, "rmse"),
                    ref_tpwcs=imcats[0],
                    clip_accum=True,
                )

        for imcat in imcats:
            image_model = imcat.meta["image_model"]()
            image_model.meta.cal_step["tweakreg"] = "COMPLETE"

            # retrieve fit status and update wcs if fit is successful:
            if "SUCCESS" in imcat.meta.get("fit_info")["status"]:
                # Update/create the WCS .name attribute with information
                # on this astrometric fit as the only record that it was
                # successful:
                if ALIGN_TO_ABS_REFCAT:
                    # NOTE: This .name attrib agreed upon by the JWST Cal
                    #       Working Group.
                    #       Current value is merely a place-holder based
                    #       on HST conventions. This value should also be
                    #       translated to the FITS WCSNAME keyword
                    #       IF that is what gets recorded in the archive
                    #       for end-user searches.
                    imcat.wcs.name = f"FIT-LVL2-{self.abs_refcat}"

                # serialize object from tweakwcs
                # (typecasting numpy objects to python types so that it doesn't cause an
                # issue when saving datamodel to ASDF)
                wcs_fit_results = {
                    k: v.tolist() if isinstance(v, (np.ndarray, np.bool_)) else v
                    for k, v in imcat.meta["fit_info"].items()
                }
                # add fit results and new WCS to datamodel
                image_model.meta["wcs_fit_results"] = wcs_fit_results
                # remove unwanted keys from WCS fit results
                for k in [
                    "eff_minobj",
                    "matched_ref_idx",
                    "matched_input_idx",
                    "fit_RA",
                    "fit_DEC",
                    "fitmask",
                ]:
                    del image_model.meta["wcs_fit_results"][k]

                image_model.meta.wcs = imcat.wcs

        return images

    def _is_wcs_correction_small(self, wcs, twcs):
        """Check that the newly tweaked wcs hasn't gone off the rails"""
        tolerance = 10.0 * self.tolerance * u.arcsec

        ra, dec = wcs.footprint(axis_type="spatial").T
        tra, tdec = twcs.footprint(axis_type="spatial").T
        skycoord = SkyCoord(ra=ra, dec=dec, unit="deg")
        tskycoord = SkyCoord(ra=tra, dec=tdec, unit="deg")

        separation = skycoord.separation(tskycoord)

        return (separation < tolerance).all()

    def _imodel2wcsim(self, image_model):
        image_model = (
            image_model
            if isinstance(image_model, rdm.DataModel)
            else rdm.open(os.path.basename(image_model))
        )
        catalog = image_model.meta.tweakreg_catalog
        model_name = os.path.splitext(image_model.meta.filename)[0].strip("_- ")

        try:
            if self.use_custom_catalogs:
                catalog_format = self.catalog_format
            else:
                catalog_format = "ascii.ecsv"

            if isinstance(catalog, str):
                # a string with the name of the catalog was provided
                catalog = Table.read(catalog, format=catalog_format)
            else:
                # catalog is a structured array, convert to astropy table:
                catalog = Table(catalog)

            catalog.meta["name"] = (
                str(catalog) if isinstance(catalog, str) else model_name
            )
        except OSError:
            self.log.error(f"Cannot read catalog {catalog}")

        # make sure catalog has 'x' and 'y' columns
        for axis in ["x", "y"]:
            if axis not in catalog.colnames:
                long_axis = axis + "centroid"
                if long_axis in catalog.colnames:
                    catalog.rename_column(long_axis, axis)
                else:
                    raise ValueError(
                        "'tweakreg' source catalogs must contain either columns 'x' and"
                        " 'y' or 'xcentroid' and 'ycentroid'."
                    )

        # create WCSImageCatalog object:
        refang = image_model.meta.wcsinfo
        # TODO: create RSTWCSCorrector in tweakwcs
        im = JWSTWCSCorrector(
            wcs=image_model.meta.wcs,
            wcsinfo={
                "roll_ref": refang["roll_ref"],
                "v2_ref": refang["v2_ref"],
                "v3_ref": refang["v3_ref"],
            },
            meta={
                "image_model": weakref.ref(image_model),
                "catalog": catalog,
                "name": model_name,
            },
        )

        return im


def _common_name(group):
    file_names = []
    for im in group:
        if isinstance(im, rdm.DataModel):
            file_names.append(os.path.splitext(im.meta.filename)[0].strip("_- "))
        else:
            raise TypeError("Input must be a list of datamodels list.")

    cn = os.path.commonprefix(file_names)
    return cn


def _parse_catfile(catfile):
    if catfile is None or not catfile.strip():
        return None

    catdict = {}

    with open(catfile) as f:
        catfile_dir = os.path.dirname(catfile)

        for line in f.readlines():
            sline = line.strip()
            if not sline or sline[0] == "#":
                continue

            data_model, *catalog = sline.split()
            catalog = list(map(str.strip, catalog))
            if len(catalog) == 1:
                catdict[data_model] = os.path.join(catfile_dir, catalog[0])
            elif len(catalog) == 0:
                catdict[data_model] = None
            else:
                raise ValueError("'catfile' can contain at most two columns.")

    return catdict



import os
from io import StringIO
from typing import Tuple

import numpy as np
import pytest
import requests
from astropy import coordinates as coord
from astropy import table
from astropy import units as u
from astropy.modeling import models
from astropy.modeling.models import RotationSequence3D, Scale, Shift
from astropy.stats import mad_std
from astropy.time import Time
from gwcs import coordinate_frames as cf
from gwcs import wcs
from gwcs.geometry import CartesianToSpherical, SphericalToCartesian
from numpy.testing import assert_allclose
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils

from romancal.tweakreg.astrometric_utils import (
    compute_radius,
    create_astrometric_catalog,
    get_catalog,
)

ARAD = np.pi / 180.0


class MockConnectionError:
    def __init__(self, *args, **kwargs):
        raise requests.exceptions.ConnectionError


def get_parallax_correction_barycenter(epoch, gaia_ref_epoch_coords):
    """
    Calculates the parallax correction in the Earth barycenter frame for a given epoch
    and Gaia reference epoch coordinates (i.e. Gaia coordinates at the reference epoch).

    Parameters
    ----------
    epoch : float
        The epoch for which the parallax correction is calculated.
    gaia_ref_epoch_coords : dict
        The Gaia reference epoch coordinates, including 'ra', 'dec', and 'parallax'.

    Returns
    -------
    tuple
        A tuple containing the delta_ra and delta_dec values of the parallax correction
        in degrees.

    Examples
    --------
    .. code-block :: python
        epoch = 2022.5
        gaia_coords = {'ra': 180.0, 'dec': 45.0, 'parallax': 10.0}
        correction = get_parallax_correction_earth_barycenter(epoch, gaia_coords)
        print(correction)
        (0.001, -0.002)
    """  # noqa: E501

    obs_date = Time(epoch, format="decimalyear")
    earths_center_barycentric_coords = coord.get_body_barycentric(
        "earth", obs_date, ephemeris="builtin"
    )
    earth_X = earths_center_barycentric_coords.x
    earth_Y = earths_center_barycentric_coords.y
    earth_Z = earths_center_barycentric_coords.z

    # angular displacement components
    # (see eq. 8.15 of "Spherical Astronomy" by Robert M. Green)
    delta_ra = (
        u.Quantity(gaia_ref_epoch_coords["parallax"], unit="mas").to(u.rad)
        * (1 / np.cos(gaia_ref_epoch_coords["dec"] * ARAD))
        * (
            earth_X.value * np.sin(gaia_ref_epoch_coords["ra"] * ARAD)
            - earth_Y.value * np.cos(gaia_ref_epoch_coords["ra"] * ARAD)
        )
    ).to("deg")
    delta_dec = (
        u.Quantity(gaia_ref_epoch_coords["parallax"], unit="mas").to(u.rad)
        * (
            earth_X.value
            * np.cos(gaia_ref_epoch_coords["ra"] * ARAD)
            * np.sin(gaia_ref_epoch_coords["dec"] * ARAD)
            + earth_Y.value
            * np.sin(gaia_ref_epoch_coords["ra"] * ARAD)
            * np.sin(gaia_ref_epoch_coords["dec"] * ARAD)
            - earth_Z.value * np.cos(gaia_ref_epoch_coords["dec"] * ARAD)
        )
    ).to("deg")

    return delta_ra, delta_dec


def get_proper_motion_correction(epoch, gaia_ref_epoch_coords, gaia_ref_epoch):
    """
    Calculates the proper motion correction for a given epoch and Gaia reference epoch
    coordinates.

    Parameters
    ----------
    epoch : float
        The epoch for which the proper motion correction is calculated.
    gaia_ref_epoch_coords : dict
        A dictionary containing Gaia reference epoch coordinates.
    gaia_ref_epoch : float
        The Gaia reference epoch.

    Returns
    -------
    None

    Examples
    --------
    .. code-block:: python
    epoch = 2022.5
    gaia_coords = {
        "ra": 180.0,
        "dec": 45.0,
        "pmra": 2.0,
        "pmdec": 1.5
    }
    gaia_ref_epoch = 2020.0
    get_proper_motion_correction(epoch, gaia_coords, gaia_ref_epoch)
    """  # noqa: E501

    expected_new_dec = (
        np.array(
            gaia_ref_epoch_coords["dec"] * 3600
            + (epoch - gaia_ref_epoch) * gaia_ref_epoch_coords["pmdec"] / 1000
        )
        / 3600
    )
    average_dec = np.array(
        [
            np.mean([new, old])
            for new, old in zip(expected_new_dec, gaia_ref_epoch_coords["dec"])
        ]
    )
    pmra = gaia_ref_epoch_coords["pmra"] / np.cos(np.deg2rad(average_dec))

    # angular displacement components
    gaia_ref_epoch_coords["pm_delta_dec"] = u.Quantity(
        (epoch - gaia_ref_epoch) * gaia_ref_epoch_coords["pmdec"] / 1000,
        unit=u.arcsec,
    ).to(u.deg)
    gaia_ref_epoch_coords["pm_delta_ra"] = u.Quantity(
        (epoch - gaia_ref_epoch) * (pmra / 1000), unit=u.arcsec
    ).to(u.deg)


def get_parallax_correction(epoch, gaia_ref_epoch_coords):
    """
    Calculates the parallax correction for a given epoch and Gaia reference epoch
    coordinates.

    Parameters
    ----------
    epoch : float
        The epoch for which to calculate the parallax correction.
    gaia_ref_epoch_coords : dict
        A dictionary containing the Gaia reference epoch coordinates:
        - "ra" : float
            The right ascension in degrees.
        - "dec" : float
            The declination in degrees.
        - "parallax" : float
            The parallax in milliarcseconds (mas).

    Returns
    -------
    None

    Notes
    -----
    This function calculates the parallax correction for a given epoch and Gaia
    reference epoch coordinates. It uses the `get_parallax_correction_barycenter`
    and `get_parallax_correction_mast` functions to obtain the parallax corrections
    based on different coordinate frames.

    Examples
    --------
    This function is typically used to add parallax correction columns to a main table
    of Gaia reference epoch coordinates.

    .. code-block:: python

        epoch = 2023.5
        gaia_coords = {
            "ra": 180.0,
            "dec": 30.0,
            "parallax": 2.5
        }
        get_parallax_correction(epoch, gaia_coords)
    """  # noqa: E501

    # get parallax correction using textbook calculations (i.e. Earth's barycenter)
    parallax_corr = get_parallax_correction_barycenter(
        epoch=epoch, gaia_ref_epoch_coords=gaia_ref_epoch_coords
    )

    # add parallax corrections columns to the main table
    gaia_ref_epoch_coords["parallax_delta_ra"] = parallax_corr[0]
    gaia_ref_epoch_coords["parallax_delta_dec"] = parallax_corr[1]


def update_wcsinfo(input_dm):
    """
    Update WCSInfo with realistic data (i.e. obtained from romanisim simulations).

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    """
    input_dm.meta.wcsinfo.v2_ref = 0.42955128282521254
    input_dm.meta.wcsinfo.v3_ref = -0.2479976768255853
    input_dm.meta.wcsinfo.vparity = -1
    input_dm.meta.wcsinfo.v3yangle = -999999
    input_dm.meta.wcsinfo.ra_ref = 270.0
    input_dm.meta.wcsinfo.dec_ref = 66.0
    input_dm.meta.wcsinfo.roll_ref = 60
    input_dm.meta.wcsinfo.s_region = (
        "POLYGON ICRS "
        "269.3318903230621 65.56866666048172 "
        "269.32578768154605 65.69246311613287 "
        "269.02457173246125 65.69201346248587 "
        "269.0333096074621 65.56870823657276 "
    )


def create_wcs_for_tweakreg_pipeline(input_dm, shift_1=0, shift_2=0):
    """
    Create a basic WCS object (with optional shift) and
    append it to the input_dm.meta attribute.

    The WCS object will have a pipeline with the required
    steps to validate against the TweakReg pipeline.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    shift_1 : int, optional
        Shift to be applied in the direction of the first axis, by default 0
    shift_2 : int, optional
        Shift to be applied in the direction of the second axis, by default 0
    """

    shape = input_dm.data.shape

    # create necessary transformations
    distortion = Shift(-shift_1) & Shift(-shift_2)
    distortion.bounding_box = ((-0.5, shape[-1] + 0.5), (-0.5, shape[-2] + 0.5))
    tel2sky = _create_tel2sky_model(input_dm)

    # create required frames
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    # create pipeline
    pipeline = [
        wcs.Step(detector, distortion),
        wcs.Step(v2v3, tel2sky),
        wcs.Step(world, None),
    ]

    wcs_obj = wcs.WCS(pipeline)

    input_dm.meta["wcs"] = wcs_obj


def _create_tel2sky_model(input_dm):
    """
    Create the transform from telescope to sky.

    The transform is defined with a reference point in a Frame
    associated with the telescope (V2, V3) in arcsec, the corresponding
    reference point on sky (RA_REF, DEC_REF) in deg, and the position angle
    at the center of the aperture, ROLL_REF in deg.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.

    Returns
    -------
    astropy.modeling.core.CompoundModel
        An astropy compound model with the transformations to map
        the telescope reference system onto the world reference systems.
    """
    v2_ref = input_dm.meta.wcsinfo.v2_ref / 3600
    v3_ref = input_dm.meta.wcsinfo.v3_ref / 3600
    roll_ref = input_dm.meta.wcsinfo.roll_ref
    ra_ref = input_dm.meta.wcsinfo.ra_ref
    dec_ref = input_dm.meta.wcsinfo.dec_ref

    angles = np.array([v2_ref, -v3_ref, roll_ref, dec_ref, -ra_ref])
    axes = "zyxyz"
    rot = RotationSequence3D(angles, axes_order=axes)

    # The sky rotation expects values in deg.
    # This should be removed when models work with quantities.
    model = (
        (Scale(0.1 / 3600) & Scale(0.1 / 3600))
        | SphericalToCartesian(wrap_lon_at=180)
        | rot
        | CartesianToSpherical(wrap_lon_at=360)
    )
    model.name = "v23tosky"
    return model


def create_basic_wcs(
    img_shape: tuple = (100, 100),
    ref_pix: tuple = (0, 0),
    ref_val: Tuple[u.Quantity, u.Quantity] = (
        u.Quantity("10 deg"),
        u.Quantity("0 deg"),
    ),
    pix_scale: Tuple[u.Quantity, u.Quantity] = (
        u.Quantity("0.1 arcsec"),
        u.Quantity("0.1 arcsec"),
    ),
    theta: u.Quantity = u.Quantity("0 deg"),
):
    """
    Creates a basic WCS (no distortion) to map pixel coordinates
    onto celestial spherical coordinates.

    Parameters
    ----------
    img_shape : tuple, optional
        Dimension of the pixel coordinates, by default (100, 100).
    ref_pix : tuple, optional
        Projection plane coordinates of the reference pixel, by default (0, 0)
    ref_val : tuple[u.Quantity, u.Quantity], optional
        Celestial coordinates at the reference pixel in degrees,
        by default (10, 0)*u.degree
    pix_scale : tuple[u.Quantity, u.Quantity], optional
        Size of a pixel side in arcsec, by default (0.1, 0.1)*u.arcsec
    theta : u.Quantity, optional
        CCW rotation angle of the projection plane, by default 0*u.deg

    Returns
    -------
    type: gwcs.wcs.WCS object
        A gwcs.wcs.WCS object with a model f : (x, y) |-> (ra, dec).

    Notes
    -----
    With the default parameters, the result will be a WCS for an image
    that's 100x100 pixels (@0.1"/pix in both axis) and reference
    coordinates (ra=10 deg, dec=0 deg, ICRS) sitting at the origin of the
    pixel coordinates frame (0,0).

    The WCS object created by this method generates a pipeline that
    does not contain the required steps to validate against the
    TweakReg pipeline.
    """

    # linear transformations
    shift_pixel_coords = models.Shift(-ref_pix[0]) & models.Shift(-ref_pix[1])
    scaling_matrix = np.array(
        [[pix_scale.to("deg")[0].value, 0], [0, pix_scale.to("deg")[1].value]]
    )
    rot_matrix = np.array(
        [
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)],
        ]
    )
    matrix = np.dot(rot_matrix, scaling_matrix)
    affine_transf = models.AffineTransformation2D(matrix)
    affine_transf.inverse = models.AffineTransformation2D(np.linalg.inv(matrix))
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        ref_val[0].value, ref_val[1].value, 180
    )
    det2sky = shift_pixel_coords | affine_transf | tan | celestial_rotation
    det2sky.name = "linear_transform"

    # frames
    detector_frame = cf.Frame2D(
        name="detector", unit=(u.pix, u.pix), axes_names=("x", "y")
    )
    cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    sky_frame = cf.CelestialFrame(
        name="sky", unit=(u.deg, u.deg), reference_frame=coord.ICRS()
    )

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]
    wcsobj = wcs.WCS(pipeline)
    wcsobj.bounding_box = (
        (-0.5, img_shape[0] + 0.5),
        (-0.5, img_shape[1] + 0.5),
    )

    return wcsobj


@pytest.fixture
def base_image():
    """
    Create a base image with a realistic WCSInfo and a WCS.

    Notes
    -----
    The size of the image needs to be relatively large in order for
    the source catalog step to find a reasonable number of sources in the image.

    shift_1 and shift_2 (units in pixel) are used to shift the WCS projection plane.
    """

    def _base_image(shift_1=0, shift_2=0):
        l2 = maker_utils.mk_level2_image(shape=(2000, 2000))
        l2.meta.exposure.mid_time = Time("2016-01-01T00:00:00")
        # update wcsinfo
        update_wcsinfo(l2)
        # add a dummy WCS object
        create_wcs_for_tweakreg_pipeline(l2, shift_1=shift_1, shift_2=shift_2)
        l2_im = rdm.ImageModel(l2)
        return l2_im

    return _base_image


@pytest.mark.parametrize(
    "catalog, num_sources",
    [
        ("GAIADR1", 5),
        ("GAIADR2", 10),
        ("GAIADR3", 15),
    ],
)
def test_create_astrometric_catalog_variable_num_sources(
    tmp_path, catalog, num_sources, request
):
    """Test fetching data from supported catalogs with variable number of sources."""
    output_filename = "ref_cat.ecsv"
    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    res = create_astrometric_catalog(
        [img],
        catalog=catalog,
        output=os.path.join(tmp_path, output_filename),
        num_sources=num_sources,
    )

    assert len(res) == num_sources


def test_create_astrometric_catalog_write_results_to_disk(tmp_path, base_image):
    img = base_image(shift_1=1000, shift_2=1000)
    num_sources = 5
    output_filename = "output"

    # get list of supported write formats
    fh = StringIO()
    table.Table.write.list_formats(out=fh)
    fh.seek(0)
    list_of_supported_formats = [
        x.strip().split()[0]
        for x in fh.readlines()[2:]
        if x.strip().split()[1].lower() == "yes"
    ]
    # exclude data formats
    [
        list_of_supported_formats.remove(x)
        for x in [
            "asdf",
            "fits",
            "hdf5",
            "parquet",
            "pandas.html",
            "pandas.json",
            "pandas.csv",
        ]
    ]

    for table_format in list_of_supported_formats:
        res = create_astrometric_catalog(
            [img],
            catalog="GAIADR3",
            output=os.path.join(tmp_path, output_filename),
            table_format=table_format,
            num_sources=num_sources,
        )

        assert len(res) == num_sources
        assert os.path.exists(os.path.join(tmp_path, output_filename))


@pytest.mark.parametrize(
    "catalog, epoch",
    [
        ("GAIADR1", "2000.0"),
        ("GAIADR2", "2010"),
        ("GAIADR3", "2030.0"),
        ("GAIADR3", "J2000"),
        ("GAIADR3", 2030.0),
        ("GAIADR3", None),
    ],
)
def test_create_astrometric_catalog_using_epoch(tmp_path, catalog, epoch, request):
    """Test fetching data from supported catalogs for a specific epoch."""
    output_filename = "ref_cat.ecsv"
    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)

    metadata_epoch = (
        epoch if epoch is not None else img.meta.exposure.mid_time.decimalyear
    )
    metadata_epoch = float(
        "".join(c for c in str(metadata_epoch) if c == "." or c.isdigit())
    )

    res = create_astrometric_catalog(
        [img],
        catalog=catalog,
        output=os.path.join(tmp_path, output_filename),
        epoch=epoch,
    )

    assert np.equal(res["epoch"], float(metadata_epoch)).all()


def test_compute_radius():
    """
    Test that compute_radius returns correct values for footprint radius and fiducial.
    """
    # create a "header" of an image that's 100x100 pixels (@0.1"/pix in both axis)
    # and reference coordinates (ra=10 deg, dec=0 deg)
    # sitting at the origin of the pixel coordinates frame (0,0).
    img_shape = (100, 100)
    ref_val = (10, 0) * u.degree
    ref_pix = (0, 0)
    pix_scale = (0.1, 0.1) * u.arcsec
    theta = 0 * u.degree

    wcsobj = create_basic_wcs(
        img_shape=img_shape,
        ref_val=ref_val,
        ref_pix=ref_pix,
        pix_scale=pix_scale,
        theta=theta,
    )

    computed_radius, computed_fiducial = compute_radius(wcsobj)

    # expected radius is the euclidean distance
    # from the center to the furthest edge of the WCS
    ref_initial_coord = (
        np.array(
            [
                (wcsobj.bounding_box[0][0]),
                (wcsobj.bounding_box[1][0]),
            ]
        )
        * pix_scale.to("deg")[0].value
    )
    fiducial_coord = (
        np.array(
            [
                (wcsobj.bounding_box[0][0] + wcsobj.bounding_box[0][1]) / 2,
                (wcsobj.bounding_box[1][0] + wcsobj.bounding_box[1][1]) / 2,
            ]
        )
        * pix_scale.to("deg")[0].value
    )
    expected_radius = np.linalg.norm(fiducial_coord - ref_initial_coord)
    # expected fiducial is the center of the WCS
    expected_fiducial = wcsobj(img_shape[0] / 2, img_shape[1] / 2)

    assert_allclose(expected_radius, computed_radius)
    assert_allclose(expected_fiducial, computed_fiducial)


@pytest.mark.parametrize(
    "ra, dec, sr, catalog_name",
    [
        (10, 10, 0.1, "GAIADR1"),
        (10, 10, 0.1, "GAIADR2"),
        (10, 10, 0.1, "GAIADR3"),
        (10, -10, 0.1, "GAIADR1"),
        (10, -10, 0.1, "GAIADR2"),
        (10, -10, 0.1, "GAIADR3"),
        (0, 0, 0.01, "GAIADR1"),
        (0, 0, 0.01, "GAIADR2"),
        (0, 0, 0.01, "GAIADR3"),
    ],
)
def test_get_catalog_using_valid_parameters(ra, dec, sr, catalog_name):
    """Test that get_catalog works properly with valid input parameters."""

    result = get_catalog(ra, dec, sr=sr, catalog=catalog_name)

    assert len(result) > 0


@pytest.mark.parametrize(
    "ra, dec, sr, catalog_name",
    [
        (10, 10, 0.1, "GAIDR3"),
        (-10, 10, 0.1, "GAIADR3"),
        (10, 100, 0.1, "GAIADR3"),
        (10, 100, 0.1, ""),
        (None, 100, 0.1, "GAIADR3"),
    ],
)
def test_get_catalog_using_invalid_parameters(ra, dec, sr, catalog_name):
    """Test that get_catalog catches all exceptions when invalid input is provided."""

    with pytest.raises(Exception) as exec_info:
        get_catalog(ra, dec, sr=sr, catalog=catalog_name)

    assert exec_info.typename.lower() == "exception"


def test_get_catalog_valid_parameters_but_no_sources_returned():
    """Test that get_catalog raises an exception if no sources are found."""

    with pytest.raises(Exception) as exec_info:
        # use a small search radius (0.5 arcsec) to force no source detection
        get_catalog(10, 10, sr=0.00014, catalog="GAIADR3")

    assert exec_info.typename.lower() == "exception"


@pytest.mark.parametrize(
    "ra, dec, epoch",
    [
        (10, 10, 2000),
        (10, 10, 2010.3),
        (10, 10, 2030),
        (10, -10, 2000),
        (10, -10, 2010.3),
        (10, -10, 2030),
        (0, 0, 2000),
        (0, 0, 2010.3),
        (0, 0, 2030),
    ],
)
def test_get_catalog_using_epoch(ra, dec, epoch):
    """Test that get_catalog returns coordinates corrected by proper motion
    and parallax. The idea is to fetch data for a specific epoch from the MAST VO API
    and compare them with the expected coordinates for that epoch.
    First, the data for a specific coordinates and epoch are fetched from the MAST VO
    API. Then, the data for the same coordinates are fetched for the Gaia's reference
    epoch of 2016.0, and corrected for proper motion and parallax using explicit
    calculations for the initially specified epoch. We then compare the results between
    the returned coordinates from the MAST VO API and the manually updated
    coordinates."""

    result = get_catalog(ra, dec, epoch=epoch)

    # updated coordinates at the provided epoch
    returned_ra = np.array(result["ra"])
    returned_dec = np.array(result["dec"])

    # get GAIA data and update coords to requested epoch using pm measurements
    gaia_ref_epoch = 2016.0
    gaia_ref_epoch_coords_all = get_catalog(ra, dec, epoch=gaia_ref_epoch)

    gaia_ref_epoch_coords = gaia_ref_epoch_coords_all  # [mask]

    # calculate proper motion corrections
    get_proper_motion_correction(
        epoch=epoch,
        gaia_ref_epoch_coords=gaia_ref_epoch_coords,
        gaia_ref_epoch=gaia_ref_epoch,
    )
    # calculate parallax corrections
    get_parallax_correction(epoch=epoch, gaia_ref_epoch_coords=gaia_ref_epoch_coords)

    # calculate the expected coordinates value after corrections have been applied to
    # Gaia's reference epoch coordinates

    # textbook (barycentric frame)
    expected_ra = (
        gaia_ref_epoch_coords["ra"]
        + gaia_ref_epoch_coords["pm_delta_ra"]
        + gaia_ref_epoch_coords["parallax_delta_ra"]
    )
    expected_dec = (
        gaia_ref_epoch_coords["dec"]
        + gaia_ref_epoch_coords["pm_delta_dec"]
        + gaia_ref_epoch_coords["parallax_delta_dec"]
    )

    assert len(result) > 0

    # adopted tolerance: 2.8e-9 deg -> 10 uas (~0.0001 pix)
    assert np.median(returned_ra - expected_ra) < 2.8e-9
    assert np.median(returned_dec - expected_dec) < 2.8e-9

    assert mad_std(returned_ra - expected_ra) < 2.8e-9
    assert mad_std(returned_dec - expected_dec) < 2.8e-9


def test_get_catalog_timeout():
    """Test that get_catalog can timeout."""

    with pytest.raises(Exception) as exec_info:
        for dt in np.arange(1, 0, -0.01):
            try:
                get_catalog(10, 10, sr=0.1, catalog="GAIADR3", timeout=dt)
            except requests.exceptions.ConnectionError:
                # ignore if it's a connection error instead of timeout
                pass

    assert exec_info.type == requests.exceptions.Timeout


def test_get_catalog_raises_connection_error(monkeypatch):
    """Test that get_catalog can raise a connection error."""

    monkeypatch.setattr("requests.get", MockConnectionError)

    with pytest.raises(Exception) as exec_info:
        get_catalog(10, 10, sr=0.1, catalog="GAIADR3")

    assert exec_info.type == requests.exceptions.ConnectionError



from astropy import units as u

from romancal.refpix.data import StandardView
from romancal.refpix.refpix import run_steps

from . import reference_utils


def test_run_steps_regression(datamodel, ref_pix_ref):
    regression = StandardView.from_datamodel(datamodel).data.copy()
    regression_out = reference_utils.apply_correction(
        regression, ref_pix_ref.alpha, ref_pix_ref.gamma, ref_pix_ref.zeta
    )

    result = run_steps(datamodel, ref_pix_ref, True, True, True, True)

    assert (result.data.value == regression_out).all()
    # regression_out does not return amp33 data

    # Check the units
    assert result.data.unit == u.DN
    assert result.amp33.unit == u.DN



"""General Utilities"""

import logging
from ast import literal_eval
from functools import wraps

from numpy.ma import masked

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def return_on_exception(exceptions=(Exception,), default=None):
    """Decorator to force functions raising exceptions to return a value

    Parameters
    ----------
    exceptions: (Exception(,...))
        Tuple of exceptions to catch

    default: obj
        The value to return when a specified exception occurs
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions as err:
                logger.debug(
                    "Caught exception %s in function %s, forcing return value of %s",
                    err,
                    func,
                    default,
                )
                return default

        return wrapper

    return decorator


def evaluate(value):
    """Evaluate a value

    Parameters
    ----------
    value : str
        The string to evaluate.

    Returns
    -------
    type or str
        The evaluation. If the value cannot be
        evaluated, the value is simply returned
    """
    try:
        evaled = literal_eval(value)
    except (ValueError, SyntaxError):
        evaled = value
    return evaled


def getattr_from_list(adict, attributes, invalid_values=None):
    """Retrieve value from dict using a list of attributes

    Parameters
    ----------
    adict : dict
        dict to retrieve from

    attributes : list
        List of attributes

    invalid_values : set
        A set of values that essentially mean the
        attribute does not exist.

    Returns
    -------
    (attribute, value)
        Returns the value and the attribute from
        which the value was taken.

    Raises
    ------
    KeyError
        None of the attributes are found in the dict.
    """
    if invalid_values is None:
        invalid_values = set()

    for attribute in attributes:
        try:
            result = adict[attribute]
        except KeyError:
            continue
        else:
            if result is masked:
                continue
            if result not in invalid_values:
                return attribute, result
            else:
                continue
    else:
        raise KeyError(f"Object has no attributes in {attributes}")


@return_on_exception(exceptions=(KeyError,), default=None)
def getattr_from_list_nofail(*args, **kwargs):
    """Call getattr_from_list without allows exceptions.

    If the specified exceptions are caught, return `default`
    instead.

    Parameters
    ----------
    See `getattr_from_list`
    """
    return getattr_from_list(*args, **kwargs)


def is_iterable(obj):
    """General iterator check"""
    return (
        not isinstance(obj, str)
        and not isinstance(obj, tuple)
        and hasattr(obj, "__iter__")
    )



""" Module to test rampfit with optional output

Notes
-----
A requirement for the larger mission verification project is to have
tests tied to reqirements.
"""

from pathlib import Path

import pytest
import roman_datamodels as rdm
from metrics_logger.decorators import metrics_logger

from romancal.lib.dms import log_result
from romancal.lib.suffix import replace_suffix
from romancal.ramp_fitting import RampFitStep

from .regtestdata import compare_asdf


# ##########
# Conditions
# ##########
def cond_is_asdf(requirement, model, expected_path):
    """Check that the filename has the correct file type"""
    msg = 'Testing that result file path has file type "asdf"'
    result = expected_path.exists() and expected_path.suffix == ".asdf"
    log_result(requirement, msg, result)
    assert result, msg


def cond_is_imagemodel(requirement, model, expected_path):
    """Check that the result is an ImageModel"""
    msg = "Testing that the result model is Level 2"
    result = isinstance(model, rdm.datamodels.ImageModel)
    log_result(requirement, msg, result)
    assert result, msg


def cond_is_rampfit(requirement, model, expected_path):
    """Check that the calibration suffix is 'rampfit'"""
    msg = 'Testing that the result file has the suffix "rampfit"'
    result = expected_path.exists() and expected_path.stem.endswith("rampfit")
    log_result(requirement, msg, result)
    assert result, msg


def cond_is_step_complete(requirement, model, expected_path):
    """Check that the calibration step is marked complete"""
    msg = "Testing that RampFitStep completed"
    result = model.meta.cal_step.ramp_fit == "COMPLETE"
    log_result(requirement, msg, result)
    assert result, msg


def cond_is_truncated(requirement, model, expected_path):
    """Check if the data represents a truncated MA table/read pattern"""
    msg = "Testing if data represents a truncated MA table"
    result = model.meta.exposure.truncated
    log_result(requirement, msg, result)
    assert result, msg


def cond_is_uneven(requirement, model, expected_path):
    """Verify that the provided model represents uneven ramps"""
    msg = "Testing that the ramps are uneven"
    length_set = {len(resultant) for resultant in model.meta.exposure.read_pattern}
    result = len(length_set) > 1
    log_result(requirement, msg, result)
    assert result, msg


def cond_science_verification(
    requirement, model, expected_path, rtdata_module, ignore_asdf_paths
):
    """Check against expected data results"""
    msg = "Testing science veracity"
    diff = compare_asdf(rtdata_module.output, rtdata_module.truth, **ignore_asdf_paths)

    result = diff.identical
    log_result(requirement, msg, result)
    assert result, diff.report()


CONDITIONS_FULL = [
    cond_is_asdf,
    cond_is_imagemodel,
    cond_is_rampfit,
    cond_is_step_complete,
    cond_is_uneven,
]
CONDITIONS_TRUNC = CONDITIONS_FULL + [cond_is_truncated]


# ######################
# fixtures and utilities
# ######################
@pytest.fixture(
    scope="module",
    params=[
        (
            "DMS362",
            Path("WFI/image/r0000101001001001001_01101_0001_WFI01_darkcurrent.asdf"),
            CONDITIONS_FULL,
        ),
        (
            "DMS366",
            Path("WFI/grism/r0000201001001001001_01101_0001_WFI01_darkcurrent.asdf"),
            CONDITIONS_FULL,
        ),
        (
            "DMS363",
            Path("WFI/image/r0000101001001001001_01101_0003_WFI01_darkcurrent.asdf"),
            CONDITIONS_TRUNC,
        ),
        (
            "DMS367",
            Path("WFI/grism/r0000201001001001001_01101_0003_WFI01_darkcurrent.asdf"),
            CONDITIONS_TRUNC,
        ),
    ],
    ids=["image_full", "spec_full", "image_trunc", "spec_trunc"],
)
def rampfit_result(request, rtdata_module):
    """Run RampFitStep

    Parameters
    ----------
    rtdata_module : pytest.fixture
        artifactory fixture for data retrieval

    Returns
    -------
    model, path : `ImageModel`, `pathlib.Path`
        Model and path to model.
    """
    # Setup inputs
    requirement, artifactory_path, conditions = request.param
    input_data = rtdata_module.get_data(str(artifactory_path))
    rtdata_module.input = input_data

    # Execute the step
    result_model = RampFitStep.call(input_data, save_results=True)

    # Setup outputs
    input_data_path = Path(input_data)
    output = replace_suffix(input_data_path.stem, "rampfit") + ".asdf"
    expected_path = input_data_path.parent / output

    # Get truths
    rtdata_module.output = output
    output_artifactory_path = Path("truth") / artifactory_path.parent / output
    rtdata_module.get_truth(str(output_artifactory_path))

    try:
        yield requirement, result_model, expected_path, conditions
    finally:
        result_model.close()


def passfail(bool_expr):
    """set pass fail"""
    if bool_expr:
        return "Pass"
    return "Fail"


# #####
# Tests
# #####
@pytest.mark.bigdata
def test_rampfit_step(rampfit_result, rtdata_module, ignore_asdf_paths):
    """Test rampfit result against various conditions"""
    requirement, model, expected_path, conditions = rampfit_result
    error_msgs = []
    for condition in conditions:
        try:
            condition(requirement, model, expected_path)
        except AssertionError as e:
            error_msgs.append(str(e))

    # Always do a full regression check.
    try:
        cond_science_verification(
            requirement, model, expected_path, rtdata_module, ignore_asdf_paths
        )
    except AssertionError as e:
        error_msgs.append(str(e))

    @metrics_logger(requirement)
    def test_success():
        assert not len(error_msgs), "\n".join(error_msgs)

    test_success()
